# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['kl']

import jax.numpy as jn

from objax.typing import JaxArray


def kl(p: JaxArray, q: JaxArray, eps: float = 2 ** -17) -> JaxArray:
    """Calculates the Kullback-Leibler divergence between arrays p and q."""
    return p.dot(jn.log(p + eps) - jn.log(q + eps))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['cross_entropy_logits', 'cross_entropy_logits_sparse', 'l2', 'sigmoid_cross_entropy_logits']

from typing import Union

import jax.numpy as jn

from objax.functional.ops import logsumexp
from objax.typing import JaxArray


def cross_entropy_logits(logits: JaxArray, labels: JaxArray) -> JaxArray:
    """Computes the softmax cross-entropy loss on n-dimensional data.

    Args:
        logits: (batch, ..., #class) tensor of logits.
        labels: (batch, ..., #class) tensor of label probabilities (e.g. labels.sum(axis=-1) must be 1)

    Returns:
        (batch, ...) tensor of the cross-entropies for each entry.
    """
    return logsumexp(logits, axis=-1) - (logits * labels).sum(-1)


def cross_entropy_logits_sparse(logits: JaxArray, labels: Union[JaxArray, int]) -> JaxArray:
    """Computes the softmax cross-entropy loss.

    Args:
        logits: (batch, ..., #class) tensor of logits.
        labels: (batch, ...) integer tensor of label indexes in {0, ...,#nclass-1} or just a single integer.

    Returns:
        (batch,) tensor of the cross-entropies for each entry.
    """
    return logsumexp(logits, axis=1) - logits[jn.arange(logits.shape[0]), labels]


def l2(x: JaxArray) -> JaxArray:
    """Computes the L2 loss.

    Args:
        x: n-dimensional tensor of floats.

    Returns:
        scalar tensor containing the l2 loss of x.
    """
    return 0.5 * (x ** 2).sum()


def sigmoid_cross_entropy_logits(logits: JaxArray, labels: Union[JaxArray, int]) -> JaxArray:
    """Computes the sigmoid cross-entropy loss.

    Args:
        logits: (batch, ..., #class) tensor of logits.
        labels: (batch, ..., #class) tensor of label probabilities (e.g. labels.sum(axis=-1) must be 1)

    Returns:
        (batch, ...) tensor of the cross-entropies for each entry.
    """
    return jn.maximum(logits, 0) - logits * labels + jn.log(1 + jn.exp(-jn.abs(logits)))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module with ResNet-v2 implementation.

See https://arxiv.org/abs/1603.05027 for detail.
"""

import functools
from typing import Callable, Sequence, Union

import objax
from objax.nn import Conv2D
from objax.typing import JaxArray

__all__ = ['ResNetV2', 'ResNet18', 'ResNet34', 'ResNet50', 'ResNet101', 'ResNet152', 'ResNet200']


def conv_args(kernel_size: int, nout: int):
    """Returns list of arguments which are common to all convolutions.
    
    Args:
        kernel_size: size of convolution kernel (single number).
        nout: number of output filters.

    Returns:
        Dictionary with common convoltion arguments.
    """
    stddev = objax.functional.rsqrt(0.5 * kernel_size * kernel_size * nout)
    return dict(w_init=functools.partial(objax.random.normal, stddev=stddev),
                use_bias=False,
                padding=objax.constants.ConvPadding.SAME)


class ResNetV2Block(objax.Module):
    """ResNet v2 block with optional bottleneck."""

    def __init__(self,
                 nin: int,
                 nout: int,
                 stride: Union[int, Sequence[int]],
                 use_projection: bool,
                 bottleneck: bool,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNetV2Block instance.

        Args:
            nin: number of input filters.
            nout: number of output filters.
            stride: stride for 3x3 convolution and projection convolution in this block.
            use_projection: if True then include projection convolution into this block.
            bottleneck: if True then make bottleneck block.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        self.use_projection = use_projection
        self.activation_fn = activation_fn

        if self.use_projection:
            self.proj_conv = Conv2D(nin, nout, 1, strides=stride, **conv_args(1, nout))

        if bottleneck:
            self.norm_0 = normalization_fn(nin)
            self.conv_0 = Conv2D(nin, nout // 4, 1, strides=1, **conv_args(1, nout // 4))
            self.norm_1 = normalization_fn(nout // 4)
            self.conv_1 = Conv2D(nout // 4, nout // 4, 3, strides=stride, **conv_args(3, nout // 4))
            self.norm_2 = normalization_fn(nout // 4)
            self.conv_2 = Conv2D(nout // 4, nout, 1, strides=1, **conv_args(1, nout))
            self.layers = ((self.norm_0, self.conv_0), (self.norm_1, self.conv_1), (self.norm_2, self.conv_2))
        else:
            self.norm_0 = normalization_fn(nin)
            self.conv_0 = Conv2D(nin, nout, 3, strides=1, **conv_args(3, nout))
            self.norm_1 = normalization_fn(nout)
            self.conv_1 = Conv2D(nout, nout, 3, strides=stride, **conv_args(3, nout))
            self.layers = ((self.norm_0, self.conv_0), (self.norm_1, self.conv_1))

    def __call__(self, x: JaxArray, training: bool) -> JaxArray:
        shortcut = x

        for i, (bn_i, conv_i) in enumerate(self.layers):
            x = bn_i(x, training)
            x = self.activation_fn(x)
            if i == 0 and self.use_projection:
                shortcut = self.proj_conv(x)
            x = conv_i(x)

        return x + shortcut


class ResNetV2BlockGroup(objax.nn.Sequential):
    """Group of ResNet v2 Blocks."""

    def __init__(self,
                 nin: int,
                 nout: int,
                 num_blocks: int,
                 stride: Union[int, Sequence[int]],
                 use_projection: bool,
                 bottleneck: bool,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNetV2BlockGroup instance.

        Args:
            nin: number of input filters.
            nout: number of output filters.
            num_blocks: number of Resnet blocks in this group.
            stride: stride for 3x3 convolutions and projection convolutions in Resnet blocks.
            use_projection: if True then include projection convolution into each Resnet blocks.
            bottleneck: if True then make bottleneck blocks.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        blocks = []
        for i in range(num_blocks):
            blocks.append(
                ResNetV2Block(
                    nin=(nin if i == 0 else nout),
                    nout=nout,
                    stride=(1 if i > 0 else stride),
                    use_projection=(i == 0 and use_projection),
                    bottleneck=bottleneck,
                    normalization_fn=normalization_fn,
                    activation_fn=activation_fn))
        super().__init__(blocks)


class ResNetV2(objax.nn.Sequential):
    """Base implementation of ResNet v2 from https://arxiv.org/abs/1603.05027."""

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 blocks_per_group: Sequence[int],
                 bottleneck: bool = True,
                 channels_per_group: Sequence[int] = (256, 512, 1024, 2048),
                 group_strides: Sequence[int] = (1, 2, 2, 2),
                 group_use_projection: Sequence[bool] = (True, True, True, True),
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNetV2 instance.

        Args:
            in_channels: number of channels in the input image.
            num_classes: number of output classes.
            blocks_per_group: number of blocks in each block group.
            bottleneck: if True then use bottleneck blocks.
            channels_per_group: number of output channels for each block group.
            group_strides: strides for each block group.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        assert len(channels_per_group) == len(blocks_per_group)
        assert len(group_strides) == len(blocks_per_group)
        assert len(group_use_projection) == len(blocks_per_group)
        nin = in_channels
        nout = 64
        ops = [Conv2D(nin, nout, k=7, strides=2, **conv_args(7, 64)),
               functools.partial(objax.functional.max_pool_2d,
                                 size=3, strides=2, padding=objax.constants.ConvPadding.SAME)]
        for i in range(len(blocks_per_group)):
            nin = nout
            nout = channels_per_group[i]
            ops.append(ResNetV2BlockGroup(
                nin,
                nout,
                num_blocks=blocks_per_group[i],
                stride=group_strides[i],
                bottleneck=bottleneck,
                use_projection=group_use_projection[i],
                normalization_fn=normalization_fn,
                activation_fn=activation_fn))

        ops.extend([normalization_fn(nout),
                    activation_fn,
                    lambda x: x.mean((2, 3)),
                    objax.nn.Linear(nout,
                                    num_classes,
                                    w_init=objax.nn.init.xavier_normal)])
        super().__init__(ops)


class ResNet18(ResNetV2):
    """Implementation of ResNet v2 with 18 layers."""

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNet18 instance.
        
        Args:
            in_channels: number of channels in the input image.
            num_classes: number of output classes.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        super().__init__(in_channels=in_channels,
                         num_classes=num_classes,
                         blocks_per_group=(2, 2, 2, 2),
                         bottleneck=False,
                         channels_per_group=(64, 128, 256, 512),
                         group_use_projection=(False, True, True, True),
                         normalization_fn=normalization_fn,
                         activation_fn=activation_fn)


class ResNet34(ResNetV2):
    """Implementation of ResNet v2 with 34 layers."""

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNet34 instance.
        
        Args:
            in_channels: number of channels in the input image.
            num_classes: number of output classes.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        super().__init__(in_channels=in_channels,
                         num_classes=num_classes,
                         blocks_per_group=(3, 4, 6, 3),
                         bottleneck=False,
                         channels_per_group=(64, 128, 256, 512),
                         group_use_projection=(False, True, True, True),
                         normalization_fn=normalization_fn,
                         activation_fn=activation_fn)


class ResNet50(ResNetV2):
    """Implementation of ResNet v2 with 50 layers."""

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNet50 instance.
        
        Args:
            in_channels: number of channels in the input image.
            num_classes: number of output classes.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        super().__init__(in_channels=in_channels,
                         num_classes=num_classes,
                         blocks_per_group=(3, 4, 6, 3),
                         bottleneck=True,
                         normalization_fn=normalization_fn,
                         activation_fn=activation_fn)


class ResNet101(ResNetV2):
    """Implementation of ResNet v2 with 101 layers."""

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNet101 instance.
        
        Args:
            in_channels: number of channels in the input image.
            num_classes: number of output classes.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        super().__init__(in_channels=in_channels,
                         num_classes=num_classes,
                         blocks_per_group=(3, 4, 23, 3),
                         bottleneck=True,
                         normalization_fn=normalization_fn,
                         activation_fn=activation_fn)


class ResNet152(ResNetV2):
    """Implementation of ResNet v2 with 152 layers."""

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNet152 instance.
        
        Args:
            in_channels: number of channels in the input image.
            num_classes: number of output classes.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        super().__init__(in_channels=in_channels,
                         num_classes=num_classes,
                         blocks_per_group=(3, 8, 36, 3),
                         bottleneck=True,
                         normalization_fn=normalization_fn,
                         activation_fn=activation_fn)


class ResNet200(ResNetV2):
    """Implementation of ResNet v2 with 200 layers."""

    def __init__(self,
                 in_channels: int,
                 num_classes: int,
                 normalization_fn: Callable[..., objax.Module] = objax.nn.BatchNorm2D,
                 activation_fn: Callable[[JaxArray], JaxArray] = objax.functional.relu):
        """Creates ResNet200 instance.
        
        Args:
            in_channels: number of channels in the input image.
            num_classes: number of output classes.
            normalization_fn: module which used as normalization function.
            activation_fn: activation function.
        """
        super().__init__(in_channels=in_channels,
                         num_classes=num_classes,
                         blocks_per_group=(3, 24, 36, 3),
                         bottleneck=True,
                         normalization_fn=normalization_fn,
                         activation_fn=activation_fn)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module with VGG-19 implementation.

See https://arxiv.org/abs/1409.1556 for detail.
"""

import functools
import os
from urllib import request

import jax.numpy as jn
import numpy as np

import objax

_VGG19_URL = 'https://github.com/machrisaa/tensorflow-vgg'
_VGG19_NPY = './objax/zoo/pretrained/vgg19.npy'
_SYNSET_URL = 'https://raw.githubusercontent.com/machrisaa/tensorflow-vgg/master/synset.txt'
_SYNSET_PATH = './objax/zoo/pretrained/synset.txt'


def preprocess(x):
    bgr_mean = [103.939, 116.779, 123.68]
    red, green, blue = [x[:, i, :, :] for i in range(3)]
    return jn.stack([blue - bgr_mean[0], green - bgr_mean[1], red - bgr_mean[2]], axis=1)


def max_pool_2d(x):
    return functools.partial(objax.functional.max_pool_2d,
                             size=2, strides=2, padding=objax.constants.ConvPadding.VALID)(x)


class VGG19(objax.nn.Sequential):
    """VGG19 implementation."""

    def __init__(self, pretrained=False):
        """Creates VGG19 instance.

        Args:
            pretrained: if True load weights from ImageNet pretrained model.
        """
        if not os.path.exists(_VGG19_NPY):
            raise FileNotFoundError(
                'You must download vgg19.npy from %s and save it to %s' % (_VGG19_URL, _VGG19_NPY))
        if not os.path.exists(_SYNSET_PATH):
            request.urlretrieve(_SYNSET_URL, _SYNSET_PATH)
        self.data_dict = np.load(_VGG19_NPY, encoding='latin1', allow_pickle=True).item()
        self.pretrained = pretrained
        self.ops = self.build()
        super().__init__(self.ops)

    def build(self):
        # inputs in [0, 255]
        self.preprocess = preprocess
        self.conv1_1 = objax.nn.Conv2D(nin=3, nout=64, k=3)
        self.relu1_1 = objax.functional.relu
        self.conv1_2 = objax.nn.Conv2D(nin=64, nout=64, k=3)
        self.relu1_2 = objax.functional.relu
        self.pool1 = max_pool_2d

        self.conv2_1 = objax.nn.Conv2D(nin=64, nout=128, k=3)
        self.relu2_1 = objax.functional.relu
        self.conv2_2 = objax.nn.Conv2D(nin=128, nout=128, k=3)
        self.relu2_2 = objax.functional.relu
        self.pool2 = max_pool_2d

        self.conv3_1 = objax.nn.Conv2D(nin=128, nout=256, k=3)
        self.relu3_1 = objax.functional.relu
        self.conv3_2 = objax.nn.Conv2D(nin=256, nout=256, k=3)
        self.relu3_2 = objax.functional.relu
        self.conv3_3 = objax.nn.Conv2D(nin=256, nout=256, k=3)
        self.relu3_3 = objax.functional.relu
        self.conv3_4 = objax.nn.Conv2D(nin=256, nout=256, k=3)
        self.relu3_4 = objax.functional.relu
        self.pool3 = max_pool_2d

        self.conv4_1 = objax.nn.Conv2D(nin=256, nout=512, k=3)
        self.relu4_1 = objax.functional.relu
        self.conv4_2 = objax.nn.Conv2D(nin=512, nout=512, k=3)
        self.relu4_2 = objax.functional.relu
        self.conv4_3 = objax.nn.Conv2D(nin=512, nout=512, k=3)
        self.relu4_3 = objax.functional.relu
        self.conv4_4 = objax.nn.Conv2D(nin=512, nout=512, k=3)
        self.relu4_4 = objax.functional.relu
        self.pool4 = max_pool_2d

        self.conv5_1 = objax.nn.Conv2D(nin=512, nout=512, k=3)
        self.relu5_1 = objax.functional.relu
        self.conv5_2 = objax.nn.Conv2D(nin=512, nout=512, k=3)
        self.relu5_2 = objax.functional.relu
        self.conv5_3 = objax.nn.Conv2D(nin=512, nout=512, k=3)
        self.relu5_3 = objax.functional.relu
        self.conv5_4 = objax.nn.Conv2D(nin=512, nout=512, k=3)
        self.relu5_4 = objax.functional.relu
        self.pool5 = max_pool_2d

        self.flatten = objax.functional.flatten
        self.fc6 = objax.nn.Linear(nin=512 * 7 * 7, nout=4096)
        self.relu6 = objax.functional.relu
        self.fc7 = objax.nn.Linear(nin=4096, nout=4096)
        self.relu7 = objax.functional.relu
        self.fc8 = objax.nn.Linear(nin=4096, nout=1000)

        if self.pretrained:
            for it in self.data_dict:
                if it.startswith('conv'):
                    conv = getattr(self, it)
                    kernel, bias = self.data_dict[it]
                    conv.w = objax.TrainVar(jn.array(kernel))
                    conv.b = objax.TrainVar(jn.array(bias[:, None, None]))
                    setattr(self, it, conv)
                elif it.startswith('fc'):
                    linear = getattr(self, it)
                    kernel, bias = self.data_dict[it]
                    if it == 'fc6':
                        kernel = kernel.reshape([7, 7, 512, -1]).transpose((2, 0, 1, 3)).reshape([512 * 7 * 7, -1])
                    linear.w = objax.TrainVar(jn.array(kernel))
                    linear.b = objax.TrainVar(jn.array(bias))
                    setattr(self, it, linear)

        ops = [self.conv1_1, self.relu1_1, self.conv1_2, self.relu1_2, self.pool1,
               self.conv2_1, self.relu2_1, self.conv2_2, self.relu2_2, self.pool2,
               self.conv3_1, self.relu3_1, self.conv3_2, self.relu3_2,
               self.conv3_3, self.relu3_3, self.conv3_4, self.relu3_4, self.pool3,
               self.conv4_1, self.relu4_1, self.conv4_2, self.relu4_2,
               self.conv4_3, self.relu4_3, self.conv4_4, self.relu4_4, self.pool4,
               self.conv5_1, self.relu5_1, self.conv5_2, self.relu5_2,
               self.conv5_3, self.relu5_3, self.conv5_4, self.relu5_4, self.pool5,
               self.flatten, self.fc6, self.relu6, self.fc7, self.relu7, self.fc8]

        return ops



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['SGD']

from typing import List

from objax.module import Module, ModuleList
from objax.typing import JaxArray
from objax.variable import TrainRef, TrainVar, VarCollection


class SGD(Module):
    """Stochastic Gradient Descent (SGD) optimizer."""

    def __init__(self, vc: VarCollection):
        """Constructor for SGD optimizer.
        
        Args:
            vc: collection of variables to optimize.
        """
        self.train_vars = ModuleList(TrainRef(x) for x in vc.subset(TrainVar))

    def __call__(self, lr: float, grads: List[JaxArray]):
        """Updates variables based on SGD algorithm.
        
        Args:
            lr: the learning rate.
            grads: the gradients to apply.
        """
        assert len(grads) == len(self.train_vars), 'Expecting as many gradients as trainable variables'
        for g, p in zip(grads, self.train_vars):
            p.value -= lr * g



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['partial', 'pmax', 'pmean', 'pmin', 'psum']

import jax
from jax.lax import lax_parallel

partial = lax_parallel.partial


def pmax(x: jax.interpreters.pxla.ShardedDeviceArray, axis_name: str = 'device'):
    """Compute a multi-device reduce max on x over the device axis axis_name."""
    return lax_parallel.pmax(x, axis_name)


def pmean(x: jax.interpreters.pxla.ShardedDeviceArray, axis_name: str = 'device'):
    """Compute a multi-device reduce mean on x over the device axis axis_name."""
    return lax_parallel.pmean(x, axis_name)


def pmin(x: jax.interpreters.pxla.ShardedDeviceArray, axis_name: str = 'device'):
    """Compute a multi-device reduce min on x over the device axis axis_name."""
    return lax_parallel.pmin(x, axis_name)


def psum(x: jax.interpreters.pxla.ShardedDeviceArray, axis_name: str = 'device'):
    """Compute a multi-device reduce sum on x over the device axis axis_name."""
    return lax_parallel.psum(x, axis_name)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable

import jax.numpy as jn

from objax import Module
from objax.nn import Linear
from objax.nn.init import kaiming_normal
from objax.typing import JaxArray
from objax.variable import TrainVar, StateVar


class RNN(Module):
    """ Recurrent Neural Network (RNN) block."""

    def __init__(self,
                 nstate: int,
                 nin: int,
                 nout: int,
                 activation: Callable = jn.tanh,
                 w_init: Callable = kaiming_normal):
        """Creates an RNN instance.

        Args:
            nstate: number of hidden units.
            nin: number of input units.
            nout: number of output units.
            activation: actication function for hidden layer.
            w_init: weight initializer for RNN model weights.
        """
        self.num_inputs = nin
        self.num_outputs = nout
        self.nstate = nstate
        self.activation = activation

        # Hidden layer parameters
        self.w_xh = TrainVar(w_init((self.num_inputs, self.nstate)))
        self.w_hh = TrainVar(w_init((self.nstate, self.nstate)))
        self.b_h = TrainVar(jn.zeros(self.nstate))

        self.output_layer = Linear(self.nstate, self.num_outputs)

    def init_state(self, batch_size):
        """Initialize hidden state for input batch of size ``batch_size``."""
        self.state = StateVar(jn.zeros((batch_size, self.nstate)))

    def __call__(self, inputs: JaxArray, only_return_final=False) -> JaxArray:
        """Forward pass through RNN.

        Args:
            inputs: ``JaxArray`` with dimensions ``num_steps, batch_size, vocabulary_size``.
            only_return_final: return only the last output if ``True``, or all output otherwise.`

        Returns:
            Output tensor with dimensions ``num_steps * batch_size, vocabulary_size``.
        """
        # Dimensions: num_steps, batch_size, vocab_size
        outputs = []
        for x in inputs:
            self.state.value = self.activation(
                jn.dot(x, self.w_xh.value) +
                jn.dot(self.state.value, self.w_hh.value) +
                self.b_h.value)
            y = self.output_layer(self.state.value)
            outputs.append(y)
        if only_return_final: return outputs[-1]
        return jn.concatenate(outputs, axis=0)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['nchw', 'nhwc', 'to_png']

import io
from typing import Union

import numpy as np
from PIL import Image

from objax.typing import JaxArray


def nchw(x: Union[np.ndarray, JaxArray]) -> Union[np.ndarray, JaxArray]:
    """Converts an array in (N,H,W,C) format to (N,C,H,W) format."""
    dims = list(range(x.ndim))
    dims.insert(-2, dims.pop())
    return x.transpose(dims)


def nhwc(x: Union[np.ndarray, JaxArray]) -> Union[np.ndarray, JaxArray]:
    """Converts an array in (N,C,H,W) format to (N,H,W,C) format."""
    dims = list(range(x.ndim))
    dims.append(dims.pop(-3))
    return x.transpose(dims)


def to_png(x: np.ndarray) -> bytes:
    """Converts numpy array in (C,H,W) format into PNG format."""
    if x.dtype in (np.float64, np.float32, np.float16):
        x = np.transpose((x + 1) * 127.5, [1, 2, 0]).clip(0, 255).round().astype('uint8')
    elif x.dtype != np.uint8:
        raise ValueError('Unsupported array type, expecting float or uint8', x.dtype)
    if x.shape[2] == 1:
        x = np.broadcast_to(x, x.shape[:2] + (3,))
    with io.BytesIO() as f:
        Image.fromarray(x).save(f, 'png')
        return f.getvalue()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['BatchNorm', 'BatchNorm0D', 'BatchNorm1D', 'BatchNorm2D',
           'Conv2D', 'ConvTranspose2D', 'Dropout', 'Linear',
           'MovingAverage', 'ExponentialMovingAverage', 'Sequential',
           'SyncedBatchNorm', 'SyncedBatchNorm0D', 'SyncedBatchNorm1D', 'SyncedBatchNorm2D']

import inspect
from typing import Callable, Iterable, Tuple, Optional, Union

from jax import numpy as jn, random as jr, lax

from objax import functional, random, util
from objax.constants import ConvPadding
from objax.module import ModuleList, Module
from objax.nn.init import kaiming_normal, xavier_normal
from objax.typing import JaxArray
from objax.variable import TrainVar, StateVar


class BatchNorm(Module):
    """Applies a batch normalization on different ranks of an input tensor.

    The module follows the operation described in Algorithm 1 of
    `Batch Normalization: Accelerating Deep Network Training by Reducing Internal Covariate Shift
    <https://arxiv.org/abs/1502.03167>`_.
    """

    def __init__(self, dims: Iterable[int], redux: Iterable[int], momentum: float = 0.999, eps: float = 1e-6):
        """Creates a BatchNorm module instance.

        Args:
            dims: shape of the batch normalization state variables.
            redux: list of indices of reduction axes. Batch norm statistics are computed by averaging over these axes.
            momentum: value used to compute exponential moving average of batch statistics.
            eps: small value which is used for numerical stability.
        """
        super().__init__()
        dims = tuple(dims)
        self.momentum = momentum
        self.eps = eps
        self.redux = tuple(redux)
        self.running_mean = StateVar(jn.zeros(dims))
        self.running_var = StateVar(jn.ones(dims))
        self.beta = TrainVar(jn.zeros(dims))
        self.gamma = TrainVar(jn.ones(dims))

    def __call__(self, x: JaxArray, training: bool) -> JaxArray:
        """Performs batch normalization of input tensor.

        Args:
            x: input tensor.
            training: if True compute batch normalization in training mode (accumulating batch statistics),
                otherwise compute in evaluation mode (using already accumulated batch statistics).

        Returns:
            Batch normalized tensor.
        """
        if training:
            m = x.mean(self.redux, keepdims=True)
            v = (x ** 2).mean(self.redux, keepdims=True) - m ** 2
            self.running_mean.value += (1 - self.momentum) * (m - self.running_mean.value)
            self.running_var.value += (1 - self.momentum) * (v - self.running_var.value)
        else:
            m, v = self.running_mean.value, self.running_var.value
        y = self.gamma.value * (x - m) * functional.rsqrt(v + self.eps) + self.beta.value
        return y


class BatchNorm0D(BatchNorm):
    """Applies a 0D batch normalization on a 2D-input batch of shape (N,C).

    The module follows the operation described in Algorithm 1 of
    `Batch Normalization: Accelerating Deep Network Training by Reducing Internal Covariate Shift
    <https://arxiv.org/abs/1502.03167>`_.
    """

    def __init__(self, nin: int, momentum: float = 0.999, eps: float = 1e-6):
        """Creates a BatchNorm0D module instance.

        Args:
            nin: number of channels in the input example.
            momentum: value used to compute exponential moving average of batch statistics.
            eps: small value which is used for numerical stability.
        """
        super().__init__((1, nin), (0,), momentum, eps)


class BatchNorm1D(BatchNorm):
    """Applies a 1D batch normalization on a 3D-input batch of shape (N,C,L).

    The module follows the operation described in Algorithm 1 of
    `Batch Normalization: Accelerating Deep Network Training by Reducing Internal Covariate Shift
    <https://arxiv.org/abs/1502.03167>`_.
    """

    def __init__(self, nin: int, momentum: float = 0.999, eps: float = 1e-6):
        """Creates a BatchNorm1D module instance.

        Args:
            nin: number of channels in the input example.
            momentum: value used to compute exponential moving average of batch statistics.
            eps: small value which is used for numerical stability.
        """
        super().__init__((1, nin, 1), (0, 2), momentum, eps)


class BatchNorm2D(BatchNorm):
    """Applies a 2D batch normalization on a 4D-input batch of shape (N,C,H,W).

    The module follows the operation described in Algorithm 1 of
    `Batch Normalization: Accelerating Deep Network Training by Reducing Internal Covariate Shift
    <https://arxiv.org/abs/1502.03167>`_.
    """

    def __init__(self, nin: int, momentum: float = 0.999, eps: float = 1e-6):
        """Creates a BatchNorm2D module instance.

        Args:
            nin: number of channels in the input example.
            momentum: value used to compute exponential moving average of batch statistics.
            eps: small value which is used for numerical stability.
        """
        super().__init__((1, nin, 1, 1), (0, 2, 3), momentum, eps)


class Conv2D(Module):
    """Applies a 2D convolution on a 4D-input batch of shape (N,C,H,W)."""

    def __init__(self,
                 nin: int,
                 nout: int,
                 k: Union[Tuple[int, int], int],
                 strides: Union[Tuple[int, int], int] = 1,
                 dilations: Union[Tuple[int, int], int] = 1,
                 groups: int = 1,
                 padding: ConvPadding = ConvPadding.SAME,
                 use_bias: bool = True,
                 w_init: Callable = kaiming_normal):
        """Creates a Conv2D module instance.

        Args:
            nin: number of channels of the input tensor.
            nout: number of channels of the output tensor.
            k: size of the convolution kernel, either tuple (height, width) or single number if they're the same.
            strides: convolution strides, either tuple (stride_y, stride_x) or single number if they're the same.
            dilations: spacing between kernel points (also known as astrous convolution),
                       either tuple (dilation_y, dilation_x) or single number if they're the same.
            groups: number of input and output channels group. When groups > 1 convolution operation is applied
                    individually for each group. nin and nout must both be divisible by groups.
            padding: padding of the input tensor, either Padding.SAME or Padding.VALID.
            use_bias: if True then convolution will have bias term.
            w_init: initializer for convolution kernel (a function that takes in a HWIO shape and returns a 4D matrix).
        """
        super().__init__()
        assert nin % groups == 0, 'nin should be divisible by groups'
        assert nout % groups == 0, 'nout should be divisible by groups'
        self.b = TrainVar(jn.zeros((nout, 1, 1))) if use_bias else None
        self.w = TrainVar(w_init((*util.to_tuple(k, 2), nin // groups, nout)))  # HWIO
        self.padding = padding
        self.strides = util.to_tuple(strides, 2)
        self.dilations = util.to_tuple(dilations, 2)
        self.groups = groups

    def __call__(self, x: JaxArray) -> JaxArray:
        """Returns the results of applying the convolution to input x."""
        y = lax.conv_general_dilated(x, self.w.value, self.strides, self.padding.value,
                                     rhs_dilation=self.dilations,
                                     feature_group_count=self.groups,
                                     dimension_numbers=('NCHW', 'HWIO', 'NCHW'))
        if self.b:
            y += self.b.value
        return y


class ConvTranspose2D(Conv2D):
    """Applies a 2D transposed convolution on a 4D-input batch of shape (N,C,H,W).

    This module can be seen as a transformation going in the opposite direction of a normal convolution, i.e.,
    from something that has the shape of the output of some convolution to something that has the shape of its input
    while maintaining a connectivity pattern that is compatible with said convolution.
    Note that ConvTranspose2D is consistent with
    `Conv2DTranspose <https://www.tensorflow.org/api_docs/python/tf/keras/layers/Conv2DTranspose>`_,
    of Tensorflow but is not consistent with
    `ConvTranspose2D <https://pytorch.org/docs/master/generated/torch.nn.ConvTranspose2d.html>`_
    of PyTorch due to kernel transpose and padding.
    """

    def __init__(self,
                 nin: int,
                 nout: int,
                 k: Union[Tuple[int, int], int],
                 strides: Union[Tuple[int, int], int] = 1,
                 dilations: Union[Tuple[int, int], int] = 1,
                 padding: ConvPadding = ConvPadding.SAME,
                 use_bias: bool = True,
                 w_init: Callable = kaiming_normal):
        """Creates a ConvTranspose2D module instance.

        Args:
            nin: number of channels of the input tensor.
            nout: number of channels of the output tensor.
            k: size of the convolution kernel, either tuple (height, width) or single number if they're the same.
            strides: convolution strides, either tuple (stride_y, stride_x) or single number if they're the same.
            dilations: spacing between kernel points (also known as astrous convolution),
                       either tuple (dilation_y, dilation_x) or single number if they're the same.
            padding: padding of the input tensor, either Padding.SAME or Padding.VALID.
            use_bias: if True then convolution will have bias term.
            w_init: initializer for convolution kernel (a function that takes in a HWIO shape and returns a 4D matrix).
        """
        super().__init__(nin=nout, nout=nin, k=k, strides=strides, padding=padding, use_bias=False, w_init=w_init)
        self.b = TrainVar(jn.zeros((nout, 1, 1))) if use_bias else None
        self.dilations = util.to_tuple(dilations, 2)

    def __call__(self, x: JaxArray) -> JaxArray:
        """Returns the results of applying the transposed convolution to input x."""
        y = lax.conv_transpose(x, self.w.value, self.strides, self.padding.value,
                               rhs_dilation=self.dilations,
                               dimension_numbers=('NCHW', 'HWIO', 'NCHW'), transpose_kernel=True)
        if self.b:
            y += self.b.value
        return y


class Dropout(Module):
    """In the training phase, a dropout layer zeroes some elements of the input tensor with probability 1-keep and
    scale the other elements by a factor of 1/keep."""

    def __init__(self, keep: float, generator=random.DEFAULT_GENERATOR):
        """Creates Dropout module instance.

        Args:
            keep: probability to keep element of the tensor.
            generator: optional argument with instance of ObJAX random generator.
        """
        self.keygen = generator
        self.keep = keep

    def __call__(self, x: JaxArray, training: bool, dropout_keep: Optional[float] = None) -> JaxArray:
        """Performs dropout of input tensor.

        Args:
            x: input tensor.
            training: if True then apply dropout to the input, otherwise keep input tensor unchanged.
            dropout_keep: optional argument, when set overrides dropout keep probability.

        Returns:
            Tensor with applied dropout.
        """
        keep = dropout_keep or self.keep
        if not training or keep >= 1:
            return x
        keep_mask = jr.bernoulli(self.keygen(), keep, x.shape)
        return jn.where(keep_mask, x / keep, 0)


class Linear(Module):
    """Applies a linear transformation on an input batch."""

    def __init__(self, nin: int, nout: int, use_bias: bool = True, w_init: Callable = xavier_normal):
        """Creates a Linear module instance.

        Args:
            nin: number of channels of the input tensor.
            nout: number of channels of the output tensor.
            use_bias: if True then linear layer will have bias term.
            w_init: weight initializer for linear layer (a function that takes in a IO shape and returns a 2D matrix).
        """
        super().__init__()
        self.b = TrainVar(jn.zeros(nout)) if use_bias else None
        self.w = TrainVar(w_init((nin, nout)))

    def __call__(self, x: JaxArray) -> JaxArray:
        """Returns the results of applying the linear transformation to input x."""
        y = jn.dot(x, self.w.value)
        if self.b:
            y += self.b.value
        return y


class MovingAverage(Module):
    """Computes moving average of an input batch."""

    def __init__(self, shape: Tuple[int, ...], buffer_size: int, init_value: float = 0):
        """Creates a MovingAverage module instance.

        Args:
            shape: shape of the input tensor.
            buffer_size: buffer size for moving average.
            init_value: initial value for moving average buffer.
        """
        self.buffer = StateVar(jn.zeros((buffer_size,) + shape) + init_value)

    def __call__(self, x: JaxArray) -> JaxArray:
        """Update the statistics using x and return the moving average."""
        self.buffer.value = jn.concatenate([self.buffer.value[1:], x[None]])
        return self.buffer.value.mean(0)


class ExponentialMovingAverage(Module):
    """computes exponential moving average (also called EMA or EWMA) of an input batch."""

    def __init__(self, shape: Tuple[int, ...], momentum: float = 0.999, init_value: float = 0):
        """Creates a ExponentialMovingAverage module instance.

        Args:
            shape: shape of the input tensor.
            momentum: momentum for exponential decrease of accumulated value.
            init_value: initial value for exponential moving average.
        """
        self.momentum = momentum
        self.avg = StateVar(jn.zeros(shape) + init_value)

    def __call__(self, x: JaxArray) -> JaxArray:
        """Update the statistics using x and return the exponential moving average."""
        self.avg.value += (self.avg.value - x) * (self.momentum - 1)
        return self.avg.value


class Sequential(ModuleList):
    """Executes modules in the order they were passed to the constructor."""

    def __call__(self, x: JaxArray, **kwargs) -> JaxArray:
        """Execute the sequence of operation contained on x and **kwargs and return result."""
        for f in self:
            s = inspect.signature(f).parameters
            if s:
                if next(reversed(s.values())).kind == inspect.Parameter.VAR_KEYWORD:
                    local_kwargs = kwargs
                else:
                    local_kwargs = {k: v for k, v in kwargs.items() if k in s}
            else:
                local_kwargs = {}
            x = f(x, **local_kwargs)
        return x


class SyncedBatchNorm(BatchNorm):
    """Synchronized batch normalization which aggregates batch statistics across all devices (GPUs/TPUs)."""

    def __call__(self, x: JaxArray, training: bool, batch_norm_update: bool = True) -> JaxArray:
        if training:
            m = functional.parallel.pmean(x.mean(self.redux, keepdims=True))
            v = functional.parallel.pmean((x ** 2).mean(self.redux, keepdims=True) - m ** 2)
            if batch_norm_update:
                self.running_mean.value += (1 - self.momentum) * (m - self.running_mean.value)
                self.running_var.value += (1 - self.momentum) * (v - self.running_var.value)
        else:
            m, v = self.running_mean.value, self.running_var.value
        y = self.gamma.value * (x - m) * functional.rsqrt(v + self.eps) + self.beta.value
        return y


class SyncedBatchNorm0D(SyncedBatchNorm):
    """Applies a 0D synchronized batch normalization on a 2D-input batch of shape (N,C).

    Synchronized batch normalization aggregated batch statistics across all devices (GPUs/TPUs) on each call.
    Compared to regular batch norm this usually leads to better accuracy at a slight performance cost.
    """

    def __init__(self, nin: int, momentum: float = 0.999, eps: float = 1e-6):
        """Creates a SyncedBatchNorm0D module instance.

        Args:
            nin: number of channels in the input example.
            momentum: value used to compute exponential moving average of batch statistics.
            eps: small value which is used for numerical stability.
        """
        super().__init__((1, nin), (0,), momentum, eps)


class SyncedBatchNorm1D(SyncedBatchNorm):
    """Applies a 1D synchronized batch normalization on a 3D-input batch of shape (N,C,L).

    Synchronized batch normalization aggregated batch statistics across all devices (GPUs/TPUs) on each call.
    Compared to regular batch norm this usually leads to better accuracy at a slight performance cost.
    """

    def __init__(self, nin: int, momentum: float = 0.999, eps: float = 1e-6):
        """Creates a SyncedBatchNorm1D module instance.

        Args:
            nin: number of channels in the input example.
            momentum: value used to compute exponential moving average of batch statistics.
            eps: small value which is used for numerical stability.
        """
        super().__init__((1, nin, 1), (0, 2), momentum, eps)


class SyncedBatchNorm2D(SyncedBatchNorm):
    """Applies a 2D synchronized batch normalization on a 4D-input batch of shape (N,C,H,W).

    Synchronized batch normalization aggregated batch statistics across all devices (GPUs/TPUs) on each call.
    Compared to regular batch norm this usually leads to better accuracy at a slight performance cost.
    """

    def __init__(self, nin: int, momentum: float = 0.999, eps: float = 1e-6):
        """Creates a SyncedBatchNorm2D module instance.

        Args:
            nin: number of channels in the input example.
            momentum: value used to compute exponential moving average of batch statistics.
            eps: small value which is used for numerical stability.
        """
        super().__init__((1, nin, 1, 1), (0, 2, 3), momentum, eps)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable, Iterable

from objax.nn import Linear, Sequential


class DNNet(Sequential):
    """Deep neural network (MLP) implementation."""

    def __init__(self, layer_sizes: Iterable[int], activation: Callable):
        """Creates DNNet instance.

        Args:
            layer_sizes: number of neurons for each layer.
            activation: layer activation.
        """
        layer_sizes = list(layer_sizes)
        assert len(layer_sizes) >= 2
        ops = []
        for i in range(1, len(layer_sizes)):
            ops.extend([Linear(layer_sizes[i - 1], layer_sizes[i]), activation])
        super().__init__(ops)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['EasyDict', 'args_indexes', 'dummy_context_mgr', 'ilog2', 'map_to_device', 'positional_args_names',
           'to_tuple']

import contextlib
import inspect
from numbers import Number
from typing import Callable, List, Union, Tuple, Iterable

import jax
import jax.numpy as jn
import numpy as np
from jax.interpreters.pxla import ShardedDeviceArray


class EasyDict(dict):
    """Custom dictionary that allows to access dict values as attributes."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__dict__ = self


def args_indexes(f: Callable, args: Iterable[str]) -> Iterable[int]:
    """Returns the indexes of variable names of a function."""
    d = {name: i for i, name in enumerate(positional_args_names(f))}
    for name in args:
        index = d.get(name)
        if index is None:
            raise ValueError(f'Function {f} does not have argument of name {name}', (f, name))
        yield index


@contextlib.contextmanager
def dummy_context_mgr():
    """Empty Context Manager."""
    yield None


def ilog2(x: float):
    """Integer log2."""
    return int(np.ceil(np.log2(x)))


map_to_device: Callable[[List[jn.ndarray]], List[ShardedDeviceArray]] = jax.pmap(lambda x: x, axis_name='device')


def positional_args_names(f: Callable) -> List[str]:
    """Returns the ordered names of the positional arguments of a function."""
    return list(p.name for p in inspect.signature(f).parameters.values()
                if p.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD))


def to_tuple(v: Union[Tuple[Number, ...], Number, Iterable], n: int):
    """Converts input to tuple."""
    if isinstance(v, tuple):
        return v
    elif isinstance(v, Number):
        return (v,) * n
    else:
        return tuple(v)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['BaseVar', 'BaseState', 'RandomState', 'TrainRef', 'StateVar', 'TrainVar', 'VarCollection']

import abc
from contextlib import contextmanager
from typing import List, Union, Tuple, Optional, Iterable, Dict, Iterator, Callable

import jax
import jax.numpy as jn
import jax.random as jr
import numpy as np

from objax.typing import JaxArray
from objax.util import map_to_device


def reduce_mean(x: JaxArray) -> JaxArray:
    return x.mean(0)


class BaseVar(abc.ABC):
    """The abstract base class to represent objax variables."""

    def __init__(self, reduce: Optional[Callable[[JaxArray], JaxArray]]):
        """Constructor for BaseVar class.

        Args:
            reduce: a function that takes an array of shape ``(n, *dims)`` and returns one of shape ``(*dims)``. Used to
                    combine the multiple states produced in an objax.Vectorize or an objax.Parallel call.
        """
        self._reduce = reduce

    @property
    @abc.abstractmethod
    def value(self) -> JaxArray:
        raise NotImplementedError('Pure method')

    @value.setter
    @abc.abstractmethod
    def value(self, tensor: JaxArray):
        raise NotImplementedError('Pure method')

    def assign(self, tensor: JaxArray):
        """Sets the value of the variable."""
        self.value = tensor

    def reduce(self, tensors: JaxArray):
        """Method called by Parallel and Vectorize to reduce a multiple-device (or batched in case of vectoriaation)
        value to a single device."""
        if self._reduce:
            self.assign(self._reduce(tensors))


class TrainVar(BaseVar):
    """A trainable variable."""

    def __init__(self, tensor: JaxArray, reduce: Optional[Callable[[JaxArray], JaxArray]] = reduce_mean):
        """TrainVar constructor.

        Args:
            tensor: the initial value of the TrainVar.
            reduce: a function that takes an array of shape ``(n, *dims)`` and returns one of shape ``(*dims)``. Used to
                    combine the multiple states produced in an objax.Vectorize or an objax.Parallel call.
        """
        self._value = tensor
        super().__init__(reduce)

    @property
    def value(self) -> JaxArray:
        """The value is read only as a safety measure to avoid accidentally making TrainVar non-differentiable.
        You can write a value to a TrainVar by using assign."""
        return self._value

    @value.setter
    def value(self, tensor: JaxArray):
        raise ValueError('Direct assignment not allowed, use TrainRef to update a TrainVar.')

    def assign(self, tensor: JaxArray):
        self._value = tensor


class BaseState(BaseVar):
    """The abstract base class used to represent objax state variables. State variables are not trainable."""

    def reduce(self, tensors: JaxArray):
        if self._reduce:
            self.assign(self._reduce(tensors))


class TrainRef(BaseState):
    """A state variable that references a trainable variable for assignment.

     TrainRef are used by optimizers to keep references to trainable variables. This is necessary to differentiate
     them from the optimizer own training variables if any."""

    def __init__(self, ref: TrainVar):
        """TrainRef constructor.

        Args:
            ref: the TrainVar to keep the reference of.
        """
        self.ref = ref
        super().__init__(None)

    @property
    def value(self) -> JaxArray:
        """The value stored in the referenced TrainVar, it can be read or written."""
        return self.ref.value

    @value.setter
    def value(self, tensor: JaxArray):
        self.ref.assign(tensor)


class StateVar(BaseState):
    """StateVar are variables that get updated manually, and are not autmatically updated by optimizers.
    For example, the mean and variance statistics in BatchNorm are StateVar."""

    def __init__(self, tensor: JaxArray, reduce: Optional[Callable[[JaxArray], JaxArray]] = reduce_mean):
        """StateVar constructor.

        Args:
            tensor: the initial value of the StateVar.
            reduce: a function that takes an array of shape ``(n, *dims)`` and returns one of shape ``(*dims)``. Used to combine
                    the multiple states produced in an objax.Vectorize or an objax.Parallel call.
        """
        self._value = tensor
        super().__init__(reduce)

    @property
    def value(self) -> JaxArray:
        """The value stored in the StateVar, it can be read or written."""
        return self._value

    @value.setter
    def value(self, tensor: JaxArray):
        self._value = tensor


class RandomState(StateVar):
    """RandomState are variables that track the random generator state. They are meant to be used internally.
    Currently only the random.Generator module uses them."""

    def __init__(self, seed: int):
        """RandomState constructor.

        Args:
            seed: the initial seed of the random number generator.
        """
        super().__init__(jr.PRNGKey(seed), None)

    def seed(self, seed: int):
        """Sets a new random seed.

        Args:
            seed: the new initial seed of the random number generator.
        """
        self.value = jr.PRNGKey(seed)

    def split(self, n: int) -> List[JaxArray]:
        """Create multiple seeds from the current seed. This is used internally by Parallel and Vectorize to ensure
        that random numbers are different in parallel threads.

        Args:
            n: the number of seeds to generate.
        """
        values = jr.split(self.value, n + 1)
        self._value = values[0]
        return values[1:]


class VarCollection(Dict[str, BaseVar]):
    """A VarCollection is a dictionary (name, var) with some additional methods to make manipulation of collections of
    variables easy. A VarCollection is ordered by insertion order. It is the object returned by Module.vars() and used
    as input by many modules: optimizers, Jit, etc..."""

    def __add__(self, other: 'VarCollection') -> 'VarCollection':
        """Overloaded add operator to merge two VarCollections together."""
        vc = VarCollection(self)
        vc.update(other)
        return vc

    def __iter__(self) -> Iterator[BaseVar]:
        """Create an iteratator that iterates over the variables (dict values) and visit them only once.
        If a variable has two names, for example in the case of weight sharing, this iterator yields the variable only
        once."""
        seen = set()
        for v in self.values():
            if v not in seen:
                seen.add(v)
                yield v

    def __setitem__(self, key: str, value: BaseVar):
        """Overload bracket assignment to catch potential conflicts during assignment."""
        if key in self:
            raise ValueError('Name conflicts when appending to VarCollection', key)
        dict.__setitem__(self, key, value)

    def update(self, other: Union['VarCollection', Iterable[Tuple[str, BaseVar]]]):
        """Overload dict.update method to catch potential conflicts during assignment."""
        keys = set(self.keys())
        if not isinstance(other, self.__class__):
            other = list(other)
        dict.update(self, other)
        if len(self) != len(keys) + len(other):
            conflicts = sorted(keys & set(other.keys()))
            raise ValueError('Name conflicts when combining VarCollection', conflicts)

    def assign(self, tensors: List[JaxArray]):
        """Assign tensors to the variables in the VarCollection. Each variable is assigned only once and in the order
        following the iter(self) iterator.

        Args:
            tensors: the list of tensors used to update variables values.
        """
        vl = list(self)
        assert len(vl) == len(tensors), f'Failed to assign a list with {len(tensors)} tensors to variables with ' \
                                        f'length {len(vl)}.'
        for var, tensor in zip(vl, tensors):
            var.assign(tensor)

    def print(self, max_width=100):
        """Pretty print the contents of the VarCollection."""
        total = count = 0
        longest_string = max((len(x) for x in self.keys()), default=20)
        longest_string = min(max_width, max(longest_string, 20))
        for name, v in self.items():
            size = np.prod(v.value.shape) if v.value.ndim else 1
            total += size
            count += 1
            print(f'{name:{longest_string}} {size:8d} {v.value.shape}')
        print(f'{f"+Total({count})":{longest_string}} {total:8d}')

    @contextmanager
    def replicate(self):
        """A context manager to use in a with statement that replicates the variables in this collection to multiple
        devices. This is used typically prior to call to objax.Parallel, so that all variables have a copy on each
        device.
        Important: replicating also updates the random state in order to have a new one per device.
        """
        ndevices = jax.device_count()
        replicated, saved_states = [], []
        for v in self:
            if isinstance(v, RandomState):
                replicated.append(v.split(ndevices))
                saved_states.append(v.value)
            else:
                replicated.append(jn.broadcast_to(v.value, (ndevices,) + v.value.shape))
        replicated = map_to_device(replicated)
        self.assign(replicated)
        yield
        visited = set()
        saved_states.reverse()
        for k, v in self.items():
            if isinstance(v, TrainRef):
                v = v.ref
                assert not isinstance(v, TrainRef)
            if v not in visited:  # Careful not to reduce twice in case of a variable and a reference to it.
                if isinstance(v, RandomState):
                    v.assign(saved_states.pop())
                else:
                    v.reduce(v.value)
                visited.add(v)

    def subset(self, is_a: Optional[Union[type, Tuple[type, ...]]] = None,
               is_not: Optional[Union[type, Tuple[type, ...]]] = None) -> 'VarCollection':
        """Return a new VarCollection that is a filtered subset of the current collection.

        Args:
            is_a: either a variable type or a list of variables types to include.
            is_not: either a variable type or a list of variables types to exclude.
        Returns:
            A new VarCollection containing the subset of variables.
        """
        vc = VarCollection()
        if is_a and is_not:
            vc.update((name, v) for name, v in self.items() if isinstance(v, is_a) and not isinstance(v, is_not))
        elif is_a:
            vc.update((name, v) for name, v in self.items() if isinstance(v, is_a))
        elif is_not:
            vc.update((name, v) for name, v in self.items() if not isinstance(v, is_not))
        return vc

    def tensors(self, is_a: Optional[Union[type, Tuple[type, ...]]] = None) -> List[JaxArray]:
        """Return the list of values for this collection. Similarly to the assign method, each variable value is
        reported only once and in the order following the iter(self) iterator.

        Args:
            is_a: either a variable type or a list of variables types to include.
        Returns:
            A new VarCollection containing the subset of variables.
        """
        if is_a:
            return [x.value for x in self if isinstance(x, is_a)]
        return [x.value for x in self]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['GradValues']

from typing import List, Optional, Callable, Tuple

import jax

from objax.module import ModuleWrapper
from objax.typing import JaxArray
from objax.variable import BaseState, TrainVar, VarCollection


class GradValues(ModuleWrapper):
    """The GradValues module is used to compute the gradients of a function."""

    def __init__(self, f: Callable,
                 variables: Optional[VarCollection],
                 input_argnums: Optional[Tuple[int, ...]] = None):
        """Constructs an instance to compute the gradient of f w.r.t. variables.

        Args:
            f: the function for which to compute gradients.
            variables: the variables for which to compute gradients.
            input_argnums: input indexes, if any, on which to compute gradients.
        """
        variables = variables or VarCollection()
        super().__init__(variables)
        self.input_argnums = input_argnums or tuple()

        def f_func(inputs_and_train_tensors: List[JaxArray],
                   state_tensors: List[JaxArray],
                   list_args: List):
            inputs = inputs_and_train_tensors[:len(self.input_argnums)]
            train_tensors = inputs_and_train_tensors[len(self.input_argnums):]
            original_vc = self.vc.tensors()
            self.vc.subset(TrainVar).assign(train_tensors)
            self.vc.subset(BaseState).assign(state_tensors)
            for i, arg in zip(self.input_argnums, inputs):
                list_args[i] = arg
            outputs = f(*list_args)
            if not isinstance(outputs, (list, tuple)):
                outputs = [outputs]
            return_value = outputs[0], (outputs, variables.tensors(BaseState))
            self.vc.assign(original_vc)
            return return_value

        self.f = jax.grad(f_func, has_aux=True)

    def __call__(self, *args):
        """Returns the computed gradients for the first value returned by `f` and the values returned by `f`.

        Returns:
            A tuple (gradients , values of f]), where gradients is a list containing
                the input gradients, if any, followed by the variable gradients."""
        inputs = [args[i] for i in self.input_argnums]
        g, (outputs, changes) = self.f(inputs + self.vc.subset(TrainVar).tensors(),
                                       self.vc.subset(BaseState).tensors(),
                                       list(args))
        self.vc.subset(BaseState).assign(changes)
        return g, outputs



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['Jit', 'Module', 'ModuleList', 'ModuleWrapper', 'Parallel', 'Vectorize']

from types import MethodType
from typing import Optional, List, Union, Callable, Tuple

import jax
import jax.numpy as jn
from jax.interpreters.pxla import ShardedDeviceArray

from objax.typing import JaxArray
from objax.util import positional_args_names
from objax.variable import BaseState, BaseVar, RandomState, VarCollection


class Module:
    """A module is a container to associate variables and functions."""

    def vars(self, scope: str = '') -> VarCollection:
        """Collect all the variables (and their names) contained in the module and its submodules.
        Important: Variables and modules stored Python structures such as dict or list are not collected. See ModuleList
        if you need such a feature.

        Args:
            scope: string to prefix to the variable names.
        Returns:
            A VarCollection of all the variables.
        """
        vc = VarCollection()
        scope += f'({self.__class__.__name__}).'
        for k, v in self.__dict__.items():
            if isinstance(v, BaseVar):
                vc[scope + k] = v
            elif isinstance(v, Module):
                vc.update(v.vars(scope=scope + k))
        return vc

    def __call__(self, *args, **kwargs):
        """Optional module __call__ method, typically a forward pass computation for standard primitives."""
        raise NotImplementedError


class ModuleList(Module, list):
    """This is a replacement for Python's list that provides a vars() method to return all the variables that it
    contains, including the ones contained in the modules and sub-modules in it."""

    def vars(self, scope: str = '') -> VarCollection:
        """Collect all the variables (and their names) contained in the list and its submodules.

        Args:
            scope: string to prefix to the variable names.
        Returns:
            A VarCollection of all the variables.
        """
        vc = VarCollection()
        scope += f'({self.__class__.__name__})'
        for p, v in enumerate(self):
            if isinstance(v, BaseVar):
                vc[f'{scope}[{p}]'] = v
            elif isinstance(v, Module):
                vc.update(v.vars(scope=f'{scope}[{p}]'))
        return vc


class ModuleWrapper(Module):
    """Module whose sole purpose is to store a collectable VarCollection. This classs is exclusively
    used internally by Objax, for example in Jit, Vectorize and Parallel."""

    def __init__(self, vc: VarCollection):
        super().__init__()
        self.vc = VarCollection((f'({self.__class__.__name__}){k}', v) for k, v in vc.items())

    def vars(self, scope: str = '') -> VarCollection:
        """Collect all the variables (and their names) contained in the VarCollection.

        Args:
            scope: string to prefix to the variable names.
        Returns:
            A VarCollection of all the variables.
        """
        return VarCollection((scope + k, v) for k, v in self.vc.items())


class Jit(ModuleWrapper):
    """JIT (Just-In-Time) module takes a function or a module and compiles it for faster execution."""

    def __init__(self,
                 f: Union[Module, Callable],
                 vc: Optional[VarCollection] = None,
                 static_argnums: Optional[Tuple[int, ...]] = None):
        """Jit constructor.

        Args:
            f: the function or the module to compile.
            vc: the VarCollection of variables used by the function or module. This argument is equired for functions.
            static_argnums: tuple of indexes of f's input arguments to treat as static (constants)).
                A new graph is compiled for each different combination of values for such inputs.
        """
        if vc is None:
            if not isinstance(f, Module):
                raise ValueError('You must supply the VarCollection used by the function f.')
            vc = f.vars()

        super().__init__(vc)
        self._call = self.jit_local_method(f, sorted(static_argnums or ()))

    def jit_local_method(self, f, static_argnums):
        """Compiles a function or module and returns method that can be attached to self instance.

        Args:
            f: function or module to compile.
            static_argnums: indexes of the arguments to be treated as static.

        Returns:
            A method containing the compiled version of f.
        """

        def jit(tensor_list: List[JaxArray], *args):
            original_values = self.vc.tensors()
            self.vc.assign(tensor_list)
            output = f(*args), self.vc.tensors(BaseState)
            self.vc.assign(original_values)
            return output

        jitf = jax.jit(jit, static_argnums=tuple(x + 1 for x in static_argnums))

        def local_method(self, *args):
            output, changes = jitf(self.vc.tensors(), *args)
            self.vc.subset(BaseState).assign(changes)
            return output

        return MethodType(local_method, self)

    def __call__(self, *args):
        """Call the compiled version of the function or module."""
        return self._call(*args)


class Parallel(ModuleWrapper):
    """Parallel module takes a function or a module and compiles it for running on multiple devices in parallel."""

    def __init__(self,
                 f: Union[Module, Callable],
                 vc: Optional[VarCollection] = None,
                 reduce: Callable[[JaxArray], JaxArray] = jn.concatenate,
                 axis_name: str = 'device',
                 static_argnums: Optional[Tuple[int, ...]] = None):
        """Parallel constructor.

        Args:
            f: the function or the module to compile for parallelism.
            vc: the VarCollection of variables used by the function or module. This argument is required for functions.
            reduce: the function used reduce the outputs from many devices to a single device value.
            axis_name: what name to give to the device dimension, used in conjunction with objax.functional.parallel.
            static_argnums: tuple of indexes of f's input arguments to treat as static (constants)).
                A new graph is compiled for each different combination of values for such inputs.
        """
        if vc is None:
            if not isinstance(f, Module):
                raise ValueError('You must supply the VarCollection used by the function f.')
            vc = f.vars()

        super().__init__(vc)
        static_argnums = sorted(static_argnums or ())
        self.reduce = reduce
        self.ndevices = jax.device_count()
        self.static_argnums = frozenset(static_argnums)

        def pmap(tensor_list: List[ShardedDeviceArray], random_list: List[ShardedDeviceArray], *args):
            original_values = self.vc.tensors()
            self.vc.assign(tensor_list)
            self.vc.subset(RandomState).assign(random_list)
            output = f(*args), self.vc.tensors(BaseState)
            self.vc.assign(original_values)
            return output

        self._call = jax.pmap(pmap, axis_name=axis_name, static_broadcasted_argnums=[x + 2 for x in static_argnums])

    def device_reshape(self, x: JaxArray) -> JaxArray:
        """Utility to reshape an input array in order to broadcast to multiple devices."""
        return x.reshape((self.ndevices, x.shape[0] // self.ndevices) + x.shape[1:])

    def __call__(self, *args):
        """Call the compiled function or module on multiple devices in parallel.
        Important: Make sure you call this function within the scope of VarCollection.replicate() statement.
        """
        args = [x if i in self.static_argnums else self.device_reshape(x) for i, x in enumerate(args)]
        output, changes = self._call(self.vc.tensors(), self.vc.subset(RandomState).tensors(), *args)
        self.vc.subset(BaseState).assign(changes)
        return jax.tree_map(self.reduce, output)


class Vectorize(ModuleWrapper):
    """Vectorize module takes a function or a module and compiles it for running in parallel on a single device."""

    def __init__(self,
                 f: Union[Module, Callable],
                 vc: Optional[VarCollection] = None,
                 batch_axis: Tuple[Optional[int], ...] = (0,)):
        """Vectorize constructor.

        Args:
            f: the function or the module to compile for vectorization.
            vc: the VarCollection of variables used by the function or module. This argument is required for functions.
            batch_axis: tuple of int or None for each of f's input arguments: the axis to use as batch during
                vectorization. Use None to automatically broadcast.
        """
        if vc is None:
            if not isinstance(f, Module):
                raise ValueError('You must supply the VarCollection used by the function f.')
            vc = f.vars()

        super().__init__(vc)
        fargs = positional_args_names(f)
        assert len(batch_axis) >= len(fargs), f'The batched argument must be specified for all of {f} arguments {fargs}'
        self.batch_axis = batch_axis
        self.batch_axis_argnums = [(x, v) for x, v in enumerate(batch_axis) if v is not None]
        assert self.batch_axis_argnums, f'No arguments to function {f} are vectorizable'

        def vmap(tensor_list: List[JaxArray], random_list: List[JaxArray], *args):
            original_values = self.vc.tensors()
            self.vc.assign(tensor_list)
            self.vc.subset(RandomState).assign(random_list)
            output = f(*args), self.vc.tensors(BaseState)
            self.vc.assign(original_values)
            return output

        self._call = jax.vmap(vmap, (None, 0) + batch_axis)

    def __call__(self, *args):
        """Call the vectorized version of the function or module."""
        assert len(args) == len(self.batch_axis), f'Number of arguments passed {len(args)} must match ' \
                                                  f'batched {len(self.batch_axis)}'
        nsplits = args[self.batch_axis_argnums[0][0]].shape[self.batch_axis_argnums[0][1]]
        output, changes = self._call(self.vc.tensors(), [v.split(nsplits) for v in self.vc.subset(RandomState)], *args)
        for v, u in zip(self.vc.subset(BaseState), changes):
            v.reduce(u)
        return output



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['load_var_collection', 'save_var_collection']

import collections
import os
from typing import IO, BinaryIO, Union

import jax.numpy as jn
import numpy as np

from objax.variable import TrainRef, VarCollection


def load_var_collection(file: Union[str, IO[BinaryIO]], vc: VarCollection):
    """Loads values of all variables in the given variables collection from file.
    
    Values loaded from file will replace old values in the variables collection.
    If variable exists in the file, but does not exist in the variables collection it will be ignored.
    If variable exists in the variables collection, but not found in the file then exception will be raised.

    Args:
        file: filename or python file handle of the input file.
        vc: variables collection which will be loaded from file.

    Raises:
        ValueError: if variable from variables collection is not found in the input file.
    """
    do_close = isinstance(file, str)
    if do_close:
        file = open(file, 'rb')
    data = np.load(file, allow_pickle=False)
    name_index = {k: str(i) for i, k in enumerate(data['names'])}
    name_vars = collections.defaultdict(list)
    for k, v in vc.items():
        if isinstance(v, TrainRef):
            v = v.ref
        name_vars[v].append(k)
    for v, names in name_vars.items():
        for name in names:
            index = name_index.get(name)
            if index is not None:
                v.assign(jn.array(data[index]))
                break
        else:
            raise ValueError(f'Missing value for variables {names}')
    if do_close:
        file.close()


def save_var_collection(file: Union[str, IO[BinaryIO]], vc: VarCollection):
    """Saves variables collection into file.
    
    Args:
        file: filename or python file handle of the file where variables will be saved.
        vc: variables collection which will be saved into file.
    """
    do_close = isinstance(file, str)
    if do_close:
        filename, file = file, open(file + '.tmp', 'wb')  # Save to a temporary in case the job is killed while saving.
    data, names, seen = {}, [], set()
    for k, v in vc.items():
        if isinstance(v, TrainRef):
            v = v.ref
        if v not in seen:
            names.append(k)
            data[str(len(data))] = v.value
            seen.add(v)
    np.savez(file, names=np.array(names), **data)
    if do_close:
        file.close()
        os.rename(filename + '.tmp', filename)  # Atomic rename to avoid broken file (when killed while saving).



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Tuple

import numpy as np
from scipy import special

__all__ = ["apply_dp_sgd_analysis", "get_privacy_spent", "compute_rdp"]


def _log_add(logx: float, logy: float) -> float:
    """Add two numbers in the log space."""
    a, b = min(logx, logy), max(logx, logy)
    if a == -np.inf:  # adding 0
        return b
    # Use exp(a) + exp(b) = (exp(a - b) + 1) * exp(b)
    return math.log1p(math.exp(a - b)) + b  # log1p(x) = log(x + 1)


def _log_sub(logx: float, logy: float) -> float:
    """Subtract two numbers in the log space. Answer must be non-negative."""
    if logx < logy:
        raise ValueError("The result of subtraction must be non-negative.")
    if logy == -np.inf:  # subtracting 0
        return logx
    if logx == logy:
        return -np.inf  # 0 is represented as -np.inf in the log space.

    try:
        # Use exp(x) - exp(y) = (exp(x - y) - 1) * exp(y).
        # expm1(x) = exp(x) - 1
        return math.log(math.expm1(logx - logy)) + logy
    except OverflowError:
        return logx


def _compute_log_a_int(q: float, sigma: float, alpha: int) -> float:
    """Compute log(A_alpha) for integer alpha. 0 < q < 1."""

    # Initialize with 0 in the log space.
    log_a = -np.inf

    for i in range(alpha + 1):
        log_coef_i = (
                math.log(special.binom(alpha, i)) + i * math.log(q) +
                (alpha - i) * math.log(1 - q))

        s = log_coef_i + (i * i - i) / (2 * (sigma ** 2))
        log_a = _log_add(log_a, s)

    return float(log_a)


def _compute_log_a_frac(q: float, sigma: float, alpha: float) -> float:
    """Compute log(A_alpha) for fractional alpha. 0 < q < 1."""
    # The two parts of A_alpha, integrals over (-inf,z0] and [z0, +inf), are
    # initialized to 0 in the log space:
    log_a0, log_a1 = -np.inf, -np.inf
    i = 0

    z0 = sigma ** 2 * math.log(1 / q - 1) + .5

    while True:  # do ... until loop
        coef = special.binom(alpha, i)
        log_coef = math.log(abs(coef))
        j = alpha - i

        log_t0 = log_coef + i * math.log(q) + j * math.log(1 - q)
        log_t1 = log_coef + j * math.log(q) + i * math.log(1 - q)

        log_e0 = math.log(.5) + _log_erfc((i - z0) / (math.sqrt(2) * sigma))
        log_e1 = math.log(.5) + _log_erfc((z0 - j) / (math.sqrt(2) * sigma))

        log_s0 = log_t0 + (i * i - i) / (2 * (sigma ** 2)) + log_e0
        log_s1 = log_t1 + (j * j - j) / (2 * (sigma ** 2)) + log_e1

        if coef > 0:
            log_a0 = _log_add(log_a0, log_s0)
            log_a1 = _log_add(log_a1, log_s1)
        else:
            log_a0 = _log_sub(log_a0, log_s0)
            log_a1 = _log_sub(log_a1, log_s1)

        i += 1
        if max(log_s0, log_s1) < -30:
            break

    return _log_add(log_a0, log_a1)


def _compute_log_a(q: float, sigma: float, alpha: float) -> float:
    """Compute log(A_alpha) for any positive finite alpha."""
    if float(alpha).is_integer():
        return _compute_log_a_int(q, sigma, int(alpha))
    else:
        return _compute_log_a_frac(q, sigma, alpha)


def _log_erfc(x: float) -> float:
    """Compute log(erfc(x)) with high accuracy for large x."""
    try:
        return math.log(2) + special.log_ndtr(-x * 2 ** .5)
    except NameError:
        # If log_ndtr is not available, approximate as follows:
        r = special.erfc(x)
        if r == 0.0:
            # Using the Laurent series at infinity for the tail of the erfc function:
            #     erfc(x) ~ exp(-x^2-.5/x^2+.625/x^4)/(x*pi^.5)
            # To verify in Mathematica:
            #     Series[Log[Erfc[x]] + Log[x] + Log[Pi]/2 + x^2, {x, Infinity, 6}]
            return (-math.log(math.pi) / 2 - math.log(x) - x ** 2 - .5 * x ** -2 +
                    .625 * x ** -4 - 37. / 24. * x ** -6 + 353. / 64. * x ** -8)
        else:
            return math.log(r)


def _compute_delta(orders: Tuple[float, ...], rdp: Tuple[float, ...], eps: float) -> Tuple[float, float]:
    """Compute delta given a list of RDP values and target epsilon.

    Args:
      orders: An array (or a scalar) of orders.
      rdp: A list (or a scalar) of RDP guarantees.
      eps: The target epsilon.

    Returns:
      Pair of (delta, optimal_order).

    Raises:
      ValueError: If input is malformed.
    """
    orders_vec = np.atleast_1d(orders)
    rdp_vec = np.atleast_1d(rdp)

    if len(orders_vec) != len(rdp_vec):
        raise ValueError("Input lists must have the same length.")

    deltas = np.exp((rdp_vec - eps) * (orders_vec - 1))
    idx_opt = np.argmin(deltas)
    return min(deltas[idx_opt], 1.), orders_vec[idx_opt]


def _compute_eps(orders: Tuple[float, ...], rdp: Tuple[float, ...], delta: float) -> Tuple[float, float]:
    """Compute epsilon given a list of RDP values and target delta.

    Args:
      orders: An array (or a scalar) of orders.
      rdp: A list (or a scalar) of RDP guarantees.
      delta: The target delta.

    Returns:
      Pair of (eps, optimal_order).

    Raises:
      ValueError: If input is malformed.
    """
    orders_vec = np.atleast_1d(orders)
    rdp_vec = np.atleast_1d(rdp)

    if len(orders_vec) != len(rdp_vec):
        raise ValueError("Input lists must have the same length.")

    eps = rdp_vec - math.log(delta) / (orders_vec - 1)

    idx_opt = np.nanargmin(eps)  # Ignore NaNs
    return eps[idx_opt], orders_vec[idx_opt]


def _compute_rdp(q: float, sigma: float, alpha: float) -> float:
    """Compute RDP of the Sampled Gaussian mechanism at order alpha.

    Args:
      q: The sampling rate.
      sigma: The std of the additive Gaussian noise.
      alpha: The order at which RDP is computed.

    Returns:
      RDP at alpha, can be np.inf.
    """
    if q == 0:
        return 0

    if q == 1.:
        return alpha / (2 * sigma ** 2)

    if np.isinf(alpha):
        return np.inf

    return _compute_log_a(q, sigma, alpha) / (alpha - 1)


def compute_rdp(q: float, noise_multiplier: float, steps: int, orders: Tuple[float, ...]):
    """Compute RDP of the Sampled Gaussian Mechanism.

    Args:
      q: The sampling rate.
      noise_multiplier: The ratio of the standard deviation of the Gaussian noise to the l2-sensitivity of the function to which it is added.
      steps: The number of steps.
      orders: An array (or a scalar) of RDP orders.

    Returns:
      The RDPs at all orders, can be np.inf.
    """

    rdp = np.array([_compute_rdp(q, noise_multiplier, order)
                    for order in orders])

    return rdp * steps


def get_privacy_spent(orders: Tuple[float, ...], rdp: Tuple[float, ...], target_eps: float = None,
                      target_delta: float = None) -> Tuple[float, float, float]:
    """Compute delta (or eps) for given eps (or delta) from RDP values.

    Args:
      orders: An array (or a scalar) of RDP orders.
      rdp: An array of RDP values. Must be of the same length as the orders list.
      target_eps: If not None, the epsilon for which we compute the corresponding delta.
      target_delta: If not None, the delta for which we compute the corresponding epsilon. Exactly one of target_eps and target_delta must be None.

    Returns:
      eps, delta, opt_order.

    Raises:
      ValueError: If target_eps and target_delta are messed up.
    """
    if target_eps is None and target_delta is None:
        raise ValueError(
            "Exactly one out of eps and delta must be None. (Both are).")

    if target_eps is not None and target_delta is not None:
        raise ValueError(
            "Exactly one out of eps and delta must be None. (None is).")

    if target_eps is not None:
        delta, opt_order = _compute_delta(orders, rdp, target_eps)
        return target_eps, delta, opt_order
    if target_delta is not None:
        eps, opt_order = _compute_eps(orders, rdp, target_delta)
        return (eps, target_delta, opt_order)


def apply_dp_sgd_analysis(q: float, noise_multiplier: float, steps: int, orders: Tuple[float, ...] = (
        1.25, 1.5, 1.75, 2., 2.25, 2.5, 3., 3.5, 4., 4.5, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20),
                          delta: float = 1e-05) -> float:
    """Compute and print results of DP-SGD analysis.

    Args:
      q: The sampling rate.
      noise_multiplier: The ratio of the standard deviation of the Gaussian noise to the l2-sensitivity of the function to which it is added.
      steps: The number of steps.
      orders: An array (or a scalar) of RDP orders.
      delta: The target delta.

    Returns:
      eps

    Raises:
      ValueError: If target_delta are messed up.
    """
    if noise_multiplier == 0:
        return float('inf')

    rdp = compute_rdp(q, noise_multiplier, steps, orders)
    eps, _, opt_order = get_privacy_spent(orders, rdp, target_delta=delta)

    return eps



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['Adam']

from typing import List

from jax import numpy as jn

from objax import functional
from objax.module import Module, ModuleList
from objax.typing import JaxArray
from objax.variable import TrainRef, StateVar, TrainVar, VarCollection


class Adam(Module):
    """Adam optimizer."""

    def __init__(self, vc: VarCollection, beta1: float = 0.9, beta2: float = 0.999, eps: float = 1e-8):
        """Constructor for Adam optimizer class.
        
        Args:
            vc: collection of variables to optimize.
            beta1: value of Adam's beta1 hyperparameter. Defaults to 0.9.
            beta2: value of Adam's beta2 hyperparameter. Defaults to 0.999.
            eps: value of Adam's epsilon hyperparameter. Defaults to 1e-8.
        """
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self.step = StateVar(jn.array(0, jn.uint32), reduce=lambda x: x[0])
        self.train_vars = ModuleList(TrainRef(x) for x in vc.subset(TrainVar))
        self.m = ModuleList(StateVar(jn.zeros_like(x.value)) for x in self.train_vars)
        self.v = ModuleList(StateVar(jn.zeros_like(x.value)) for x in self.train_vars)

    def __call__(self, lr: float, grads: List[JaxArray]):
        """Updates variables and other state based on Adam algorithm.
        
        Args:
            lr: the learning rate.
            grads: the gradients to apply.
        """
        assert len(grads) == len(self.train_vars), 'Expecting as many gradients as trainable variables'
        self.step.value += 1
        lr *= jn.sqrt(1 - self.beta2 ** self.step.value) / (1 - self.beta1 ** self.step.value)
        for g, p, m, v in zip(grads, self.train_vars, self.m, self.v):
            m.value += (1 - self.beta1) * (g - m.value)
            v.value += (1 - self.beta2) * (g ** 2 - v.value)
            p.value -= lr * m.value * functional.rsqrt(v.value + self.eps)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['PrivateGradValues']

from typing import Optional, Callable, Tuple

import jax.numpy as jn

from objax import random
from objax.gradient import GradValues
from objax.module import Module, Vectorize
from objax.typing import JaxArray
from objax.variable import VarCollection


class PrivateGradValues(Module):
    """Computes differentially private gradients as required by DP-SGD.
    This module can be used in place of GradVals, and automatically makes the optimizer differentially private."""

    def __init__(self,
                 f: Callable,
                 vc: VarCollection,
                 noise_multiplier: float,
                 l2_norm_clip: float,
                 microbatch: int,
                 batch_axis: Tuple[Optional[int], ...] = (0,),
                 keygen: random.Generator = random.DEFAULT_GENERATOR):
        """Constructs a PrivateGradValues instance.
 
        Args:
            f: the function for which to compute gradients.
            vc: the variables for which to compute gradients.
            noise_multiplier: scale of standard deviation for added noise in DP-SGD.
            l2_norm_clip: value of clipping norm for DP-SGD.
            microbatch: the size of each microbatch.
            batch_axis: the axis to use as batch during vectorization. Should be a tuple of 0s.
            keygen: a Generator for random numbers. Defaults to objax.random.DEFAULT_GENERATOR.
        """
        super().__init__()

        if not all(v == 0 for v in batch_axis):
            raise ValueError('batch_axis needs to be an all zero tuple for PrivateGradValues.')

        gv = GradValues(f, vc)

        def clipped_grad(*args):
            grads, values = gv(*args)
            total_grad_norm = jn.linalg.norm([jn.linalg.norm(g) for g in grads])
            idivisor = 1 / jn.maximum(total_grad_norm / l2_norm_clip, 1.)
            return [g * idivisor for g in grads], values

        self.microbatch = microbatch
        self.l2_norm_clip = l2_norm_clip
        self.noise_multiplier = noise_multiplier
        self.keygen = keygen
        self.private_grad = Vectorize(clipped_grad, gv.vars(), batch_axis=batch_axis)

    def reshape_microbatch(self, x: JaxArray) -> JaxArray:
        """Reshapes examples into microbatches.
        DP-SGD requires that per-example gradients are clipped and noised, however this can be inefficient.
        To speed this up, it is possible to clip and noise a microbatch of examples, at a sight cost to privacy.
        If speed is not an issue, the microbatch size should be set to 1.

        If x has shape [D0, D1, ..., Dn], the reshaped output will
        have shape [number_of_microbatches, microbatch_size, D1, ..., DN].

        Args:
            x: items to be reshaped.

        Returns:
            The reshaped items.
        """
        s = x.shape
        return x.reshape([s[0] // self.microbatch, self.microbatch, *s[1:]])

    def __call__(self, *args):
        """Returns the computed DP-SGD gradients.

        Returns:
            A tuple (gradients, value of f)."""
        batch = args[0].shape[0]
        assert batch % self.microbatch == 0
        num_microbatches = batch // self.microbatch
        stddev = self.l2_norm_clip * self.noise_multiplier / num_microbatches
        g, v = self.private_grad(*[self.reshape_microbatch(x) for x in args])
        g, v = ([x.mean(0) for x in gv] for gv in (g, v))
        g = [gx + random.normal(gx.shape, stddev=stddev, generator=self.keygen) for gx in g]
        return g, v



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['Checkpoint']

import glob
import os
from typing import Callable, Optional

from objax.io.ops import load_var_collection, save_var_collection
from objax.typing import FileOrStr
from objax.variable import VarCollection


class Checkpoint:
    """Helper class which performs saving and restoring of the variables.
    
    Variables are stored in the checkpoint files. One checkpoint file stores a single snapshot of the variables.
    Different checkpoint files store different snapshots of the variables (for example at different training step).
    Each checkpoint has associated index, which is used to identify time when snapshot of the variables was made.
    Typically training step or training epoch are used as an index.
    """

    DIR_NAME: str = 'ckpt'
    """Name of the subdirectory of model directory where checkpoints will be saved."""

    FILE_MATCH: str = '*.npz'
    """File pattern which is used to search for checkpoint files."""

    FILE_FORMAT: str = '%010d.npz'
    """Format of the filename of one checkpoint file."""

    LOAD_FN: Callable[[FileOrStr, VarCollection], None] = staticmethod(load_var_collection)
    """Load function, which loads variables collection from given file."""

    SAVE_FN: Callable[[FileOrStr, VarCollection], None] = staticmethod(save_var_collection)
    """Save function, which saves variables collection into given file."""

    def __init__(self, logdir: str, keep_ckpts: int, makedir: bool = True, verbose: bool = True):
        """Creates instance of the Checkpoint class.

        Args:
            logdir: model directory. Checkpoints will be saved in the subdirectory of model directory.
            keep_ckpts: maximum number of checkpoints to keep.
            makedir: if True then directory for checkpoints will be created,
                otherwise it's expected that directory already exists.
            verbose: if True then print when data is restored from checkpoint.
        """
        self.logdir = logdir
        self.keep_ckpts = keep_ckpts
        self.verbose = verbose
        if makedir:
            os.makedirs(os.path.join(logdir, self.DIR_NAME), exist_ok=True)

    @staticmethod
    def checkpoint_idx(filename: str):
        """Returns index of checkpoint from given checkpoint filename.

        Args:
            filename: checkpoint filename.

        Returns:
            checkpoint index.
        """
        return int(os.path.basename(filename).split('.')[0])

    def restore(self, vc: VarCollection, idx: Optional[int] = None):
        """Restores values of all variables of given variables collection from the checkpoint.

        Old values from the variables collection will be replaced with the new values read from checkpoint.
        If variable does not exist in the variables collection, it won't be restored from checkpoint.

        Args:
            vc: variables collection to restore.
            idx: if provided then checkpoint index to use, if None then latest checkpoint will be restored.

        Returns:
            idx: index of the restored checkpoint.
            ckpt: full path to the restored checkpoint.
        """
        if idx is None:
            all_ckpts = glob.glob(os.path.join(self.logdir, self.DIR_NAME, self.FILE_MATCH))
            if not all_ckpts:
                return 0, ''
            idx = self.checkpoint_idx(max(all_ckpts))
        ckpt = os.path.join(self.logdir, self.DIR_NAME, self.FILE_FORMAT % idx)
        if self.verbose:
            print('Resuming from', ckpt)
        self.LOAD_FN(ckpt, vc)
        return idx, ckpt

    def save(self, vc: VarCollection, idx: int):
        """Saves variables collection to checkpoint with given index.
        
        Args:
            vc: variables collection to save.
            idx: index of the new checkpoint where variables should be saved.
        """
        self.SAVE_FN(os.path.join(self.logdir, self.DIR_NAME, self.FILE_FORMAT % idx), vc)
        for ckpt in sorted(glob.glob(os.path.join(self.logdir, self.DIR_NAME, self.FILE_MATCH)))[:-self.keep_ckpts]:
            os.remove(ckpt)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['DEFAULT_GENERATOR', 'Generator', 'randint', 'normal', 'truncated_normal', 'uniform']

from typing import Optional, Tuple

import jax.random as jr

from objax.module import Module
from objax.variable import RandomState, VarCollection


class Generator(Module):
    """Random number generator module."""

    def __init__(self, seed: int = 0):
        """Create a random key generator, seed is the random generator initial seed."""
        super().__init__()
        self.initial_seed = seed
        self._key: Optional[RandomState] = None

    @property
    def key(self):
        """The random generator state (a tensor of 2 int32)."""
        if self._key is None:
            self._key = RandomState(self.initial_seed)
        return self._key

    def seed(self, seed: int = 0):
        """Sets a new random generator seed."""
        self.initial_seed = seed
        if self._key is not None:
            self._key.seed(seed)

    def __call__(self):
        """Generate a new generator state."""
        return self.key.split(1)[0]

    def vars(self, scope: str = '') -> VarCollection:
        self.key  # Make sure the key is created before collecting the vars.
        return super().vars(scope)


DEFAULT_GENERATOR = Generator(0)


def normal(shape: Tuple[int, ...], *, mean: float = 0, stddev: float = 1, generator: Generator = DEFAULT_GENERATOR):
    """Returns a ``JaxArray`` of shape ``shape`` with random numbers from a normal distribution
    with mean ``mean`` and standard deviation ``stddev``."""
    return jr.normal(generator(), shape=shape) * stddev + mean


def randint(shape: Tuple[int, ...], low: int, high: int, generator: Generator = DEFAULT_GENERATOR):
    """Returns a ``JaxAarray`` of shape ``shape`` with random integers in {low, ..., high-1}."""
    return jr.randint(generator(), shape=shape, minval=low, maxval=high)


def truncated_normal(shape: Tuple[int, ...], *,
                     stddev: float = 1,
                     lower: float = -2,
                     upper: float = 2,
                     generator: Generator = DEFAULT_GENERATOR):
    """Returns a ``JaxArray`` of shape ``shape`` with random numbers from a normal distribution
    with mean 0 and standard deviation ``stddev`` truncated by (``lower``, ``upper``)."""
    return jr.truncated_normal(generator(), shape=shape, lower=lower, upper=upper) * stddev


def uniform(shape: Tuple[int, ...], generator: Generator = DEFAULT_GENERATOR):
    """Returns a ``JaxArray`` of shape ``shape`` with random numbers from a uniform distribution [0, 1]."""
    return jr.uniform(generator(), shape=shape)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['ExponentialMovingAverage']

from typing import Callable, Tuple, List

import jax.numpy as jn

from objax.module import Module, ModuleList
from objax.typing import JaxArray
from objax.variable import RandomState, TrainRef, StateVar, TrainVar, VarCollection


class ExponentialMovingAverage(Module):
    """Maintains exponential moving averages for each variable from provided VarCollection."""

    def __init__(self, vc: VarCollection, momentum: float = 0.999, debias: bool = False, eps: float = 1e-6):
        """Creates ExponentialMovingAverage instance with given hyperparameters.
        
        Args:
            momentum: the decay factor for the moving average.
            debias: bool indicating whether to use initialization bias correction.
            eps: small adjustment to prevent division by zero.
        """
        self.momentum = momentum
        self.debias = debias
        self.eps = eps
        self.step = StateVar(jn.array(0, jn.uint32), reduce=lambda x: x[0])
        # Deduplicate variables and skip RandomState vars since they cannot be averaged.
        trainable, non_trainable = {}, {}  # Use dicts since they are ordered since python >= 3.6
        for v in vc:
            if isinstance(v, RandomState):
                continue
            if isinstance(v, TrainRef):
                v = v.ref
            if isinstance(v, TrainVar):
                trainable[v] = True
            else:
                non_trainable[v] = True
        self.refs = ModuleList(list(non_trainable.keys()) + [TrainRef(v) for v in trainable.keys()])
        self.m = ModuleList(StateVar(jn.zeros_like(x.value)) for x in self.refs)

    def __call__(self):
        """Updates the moving average."""
        self.step.value += 1
        for ref, m in zip(self.refs, self.m):
            m.value += (1 - self.momentum) * (ref.value - m.value)

    def refs_and_values(self) -> Tuple[VarCollection, List[JaxArray]]:
        """Returns the VarCollection of variables affected by Exponential Moving Average (EMA) and
        their corresponding EMA values."""
        if self.debias:
            step = self.step.value
            debias = 1 / (1 - (1 - self.eps) * self.momentum ** step)
            tensors = [m.value * debias for ref, m in zip(self.refs, self.m)]
        else:
            tensors = self.m.vars().tensors()
        return self.refs.vars(), tensors

    def replace_vars(self, f: Callable):
        """Returns a function that acts as f called when variables are replaced by their averages.

        Args:
            f: function to be called on the stored averages.

        Returns:
            A function that returns the output of calling f with stored variables replaced by
            their moving averages.
        """

        def wrap(*args, **kwargs):
            refs, new_values = self.refs_and_values()
            original_values = refs.tensors()
            refs.assign(new_values)
            output = f(*args, **kwargs)
            refs.assign(original_values)
            return output

        return wrap



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['gain_leaky_relu', 'kaiming_normal', 'kaiming_normal_gain', 'kaiming_truncated_normal', 'truncated_normal',
           'xavier_normal', 'xavier_truncated_normal']

from typing import Tuple

import numpy as np
import scipy.stats

from objax import random
from objax.typing import JaxArray


# Expect format for init APIs
# Convolution: HWIO (number of output channels at the end)
# Linear: IO (number of output dimensions at the end)
# In general: the last dimensions is the number of output channels/dimensions.

def gain_leaky_relu(relu_slope: float = 0.1):
    """The recommended gain value for leaky_relu.

    Args:
        relu_slope: negative slope of leaky_relu.

    Returns:
        The recommended gain value for leaky_relu.
    """
    return np.sqrt(2 / (1 + relu_slope ** 2))


def kaiming_normal(shape: Tuple[int, ...], gain: float = 1) -> JaxArray:
    """Returns a tensor with values assigned using Kaiming He normal initializer from
    `Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification
    <https://arxiv.org/abs/1502.01852>`_.

    Args:
        shape: shape of the output tensor.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with normal random variables with standard deviation (gain * kaiming_normal_gain).
    """
    return random.normal(shape, stddev=gain * kaiming_normal_gain(shape))


def kaiming_normal_gain(shape: Tuple[int, ...]) -> float:
    """Returns Kaiming He gain from
    `Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification
    <https://arxiv.org/abs/1502.01852>`_.

    Args:
        shape: shape of the output tensor.

    Returns:
        Scalar, the standard deviation gain.
    """
    fan_in = np.prod(shape[:-1])
    return np.sqrt(1 / fan_in)


def kaiming_truncated_normal(shape: Tuple[int, ...], lower: float = -2, upper: float = 2, gain: float = 1) -> JaxArray:
    """Returns a tensor with values assigned using Kaiming He truncated normal initializer from
    `Delving Deep into Rectifiers: Surpassing Human-Level Performance on ImageNet Classification
    <https://arxiv.org/abs/1502.01852>`_.

    Args:
        shape: shape of the output tensor.
        lower: lower truncation of the normal.
        upper: upper truncation of the normal.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with truncated normal random variables with standard
        deviation (gain * kaiming_normal_gain) and support [lower, upper].
    """
    truncated_std = scipy.stats.truncnorm.std(a=lower, b=upper, loc=0., scale=1)
    stddev = gain * kaiming_normal_gain(shape) / truncated_std
    return random.truncated_normal(shape, stddev=stddev, lower=lower, upper=upper)


def xavier_normal(shape: Tuple[int, ...], gain: float = 1) -> JaxArray:
    """Returns a tensor with values assigned using Xavier Glorot normal initializer from
    `Understanding the difficulty of training deep feedforward neural networks
    <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>`_.
    
    Args:
        shape: shape of the output tensor.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with normal random variables with standard deviation (gain * xavier_normal_gain).
    """
    return random.normal(shape, stddev=gain * xavier_normal_gain(shape))


def truncated_normal(shape: Tuple[int, ...], lower: float = -2, upper: float = 2, stddev: float = 1) -> JaxArray:
    """Returns a tensor with values assigned using truncated normal initialization.

    Args:
        shape: shape of the output tensor.
        lower: lower truncation of the normal.
        upper: upper truncation of the normal.
        stddev: expected standard deviation.

    Returns:
        Tensor initialized with truncated normal random variables with standard
        deviation stddev and support [lower, upper].
    """
    truncated_std = scipy.stats.truncnorm.std(a=lower, b=upper, loc=0., scale=1)
    stddev /= truncated_std
    return random.truncated_normal(shape, stddev=stddev, lower=lower, upper=upper)


def xavier_normal_gain(shape: Tuple[int, ...]) -> float:
    """Returns Xavier Glorot gain from
    `Understanding the difficulty of training deep feedforward neural networks
    <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>`_.

    Args:
        shape: shape of the output tensor.

    Returns:
        Scalar, the standard deviation gain.
    """
    fan_in, fan_out = np.prod(shape[:-1]), shape[-1]
    return np.sqrt(2 / (fan_in + fan_out))


def xavier_truncated_normal(shape: Tuple[int, ...], lower: float = -2, upper: float = 2, gain: float = 1) -> JaxArray:
    """Returns a tensor with values assigned using Xavier Glorot truncated normal initializer from
    `Understanding the difficulty of training deep feedforward neural networks
    <http://proceedings.mlr.press/v9/glorot10a/glorot10a.pdf>`_.

    Args:
        shape: shape of the output tensor.
        lower: lower truncation of the normal.
        upper: upper truncation of the normal.
        gain: optional scaling factor.

    Returns:
        Tensor initialized with truncated normal random variables with standard
        deviation (gain * xavier_normal_gain) and support [lower, upper].
    """
    truncated_std = scipy.stats.truncnorm.std(a=lower, b=upper, loc=0., scale=1)
    stddev = gain * xavier_normal_gain(shape) / truncated_std
    return random.truncated_normal(shape, stddev=stddev, lower=lower, upper=upper)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['Momentum']

from typing import List

from jax import numpy as jn

from objax.module import Module, ModuleList
from objax.variable import TrainRef, StateVar, TrainVar, VarCollection


class Momentum(Module):
    """Momentum optimizer."""

    def __init__(self, vc: VarCollection, momentum: float = 0.9, nesterov: bool = False):
        """Constructor for momentum optimizer class.
        
        Args:
            vc: collection of variables to optimize.
            momentum: the momentum hyperparameter.
            nesterov: bool indicating whether to use the Nesterov method.
        """
        self.momentum = momentum
        self.nesterov = nesterov
        self.train_vars = ModuleList(TrainRef(x) for x in vc.subset(TrainVar))
        self.m = ModuleList(StateVar(jn.zeros_like(x.value)) for x in self.train_vars)

    def __call__(self, lr: float, grads: List[jn.ndarray]):
        """Updates variables and other state based on momentum (or Nesterov) SGD.

        Args:
           lr: the learning rate.
           grads: the gradients to apply.
        """
        assert len(grads) == len(self.train_vars), 'Expecting as many gradients as trainable variables'
        if self.nesterov:
            for g, p, m in zip(grads, self.train_vars, self.m):
                m.value = g + self.momentum * m.value
                p.value -= lr * (g + self.momentum * m.value)
        else:
            for g, p, m in zip(grads, self.train_vars, self.m):
                m.value = g + self.momentum * m.value
                p.value -= lr * m.value



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import enum
import os
from time import time
from typing import Union, Callable, Tuple, ByteString

import numpy as np
from tensorboard.compat.proto import event_pb2
from tensorboard.compat.proto import summary_pb2
from tensorboard.summary.writer.event_file_writer import EventFileWriter
from tensorboard.util.tensor_util import make_tensor_proto

from objax import util


class Reducer(enum.Enum):
    """Reduces tensor batch into a single tensor."""
    FIRST = lambda x: x[0]
    LAST = lambda x: x[-1]
    MEAN = lambda x: np.mean(x)


class DelayedScalar:
    def __init__(self, reduce: Union[Callable, Reducer]):
        self.values = []
        self.reduce = reduce

    def __call__(self):
        return self.reduce(self.values)


class Image:
    def __init__(self, shape: Tuple[int, int, int], png: ByteString):
        self.shape = shape
        self.png = png


class Text:
    def __init__(self, text: str):
        self.text = text


class Summary(dict):
    """Writes entries to `Summary` protocol buffer."""

    def image(self, tag: str, image: np.ndarray):
        """Adds image to the summary. Float image in [-1, 1] in CHW format expected."""
        self[tag] = Image(image.shape, util.image.to_png(image))

    def scalar(self, tag: str, value: float, reduce: Union[Callable, Reducer] = Reducer.MEAN):
        """Adds scalar to the summary."""
        if tag not in self:
            self[tag] = DelayedScalar(reduce)
        self[tag].values.append(value)

    def text(self, tag: str, text: str):
        """Adds text to the summary."""
        self[tag] = Text(text)

    def __call__(self):
        entries = []
        for tag, value in self.items():
            if isinstance(value, DelayedScalar):
                entries.append(summary_pb2.Summary.Value(tag=tag, simple_value=value()))
            elif isinstance(value, Image):
                image_summary = summary_pb2.Summary.Image(encoded_image_string=value.png,
                                                          colorspace=value.shape[0],
                                                          height=value.shape[1],
                                                          width=value.shape[2])
                entries.append(summary_pb2.Summary.Value(tag=tag, image=image_summary))
            elif isinstance(value, Text):
                metadata = summary_pb2.SummaryMetadata(
                    plugin_data=summary_pb2.SummaryMetadata.PluginData(plugin_name='text'))
                entries.append(summary_pb2.Summary.Value(tag=tag, metadata=metadata,
                                                         tensor=make_tensor_proto(values=value.text.encode('utf-8'),
                                                                                  shape=(1,))))
            else:
                raise NotImplementedError(tag, value)
        return summary_pb2.Summary(value=entries)


class SummaryWriter:
    """Writes entries to event files in the logdir to be consumed by Tensorboard."""

    def __init__(self, logdir: str, queue_size: int = 5, write_interval: int = 5):
        """Creates SummaryWriter instance.

        Args:
            logdir: directory where event file will be written.
            queue_size: size of the queue for pending events and summaries
                        before one of the 'add' calls forces a flush to disk.
            write_interval: how often, in seconds, to write the pending events and summaries to disk.
        """
        if not os.path.isdir(logdir):
            os.makedirs(logdir, exist_ok=True)

        self.writer = EventFileWriter(logdir, queue_size, write_interval)

    def write(self, summary: Summary, step: int):
        """Adds on event to the event file."""
        self.writer.add_event(event_pb2.Event(step=step, summary=summary(), wall_time=time()))

    def close(self):
        """Flushes the event file to disk and close the file."""
        self.writer.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['celu', 'dynamic_slice', 'elu', 'leaky_relu', 'log_sigmoid', 'log_softmax', 'logsumexp',
           'pad', 'rsqrt', 'selu', 'sigmoid', 'softmax', 'softplus', 'stop_gradient', 'tanh', 'top_k', 'relu',
           'average_pool_2d', 'flatten', 'max_pool_2d', 'one_hot', 'upscale']

from typing import Union, Tuple

import jax.nn.functions as jnnf
import jax.scipy.special
from jax import numpy as jn, lax

from objax.constants import ConvPadding
from objax.typing import JaxArray
from objax.util import to_tuple

celu = jnnf.celu
dynamic_slice = lax.dynamic_slice
elu = jnnf.elu
leaky_relu = jnnf.leaky_relu
log_sigmoid = jnnf.log_sigmoid
log_softmax = jnnf.log_softmax
logsumexp = jax.scipy.special.logsumexp
one_hot = jnnf.one_hot
pad = jn.pad
selu = jnnf.selu
sigmoid = jnnf.sigmoid
softmax = jnnf.softmax
softplus = jnnf.softplus
stop_gradient = lax.stop_gradient
tanh = lax.tanh
top_k = lax.top_k  # Current code doesn't work with gradient.
rsqrt = lax.rsqrt


def average_pool_2d(x: JaxArray,
                    size: Union[Tuple[int, int], int] = 2,
                    strides: Union[Tuple[int, int], int] = 2,
                    padding: ConvPadding = ConvPadding.VALID) -> JaxArray:
    """Applies average pooling using a square 2D filter.

    Args:
        x: input tensor of shape NCHW.
        size: size of pooling filter.
        strides: stride step.
        padding: type of padding used in pooling operation.

    Returns:
        Output tensor.
    """
    size = to_tuple(size, 2)
    strides = to_tuple(strides, 2)
    return lax.reduce_window(x, 0, lax.add, (1, 1) + size, (1, 1) + strides, padding=padding.value)


def flatten(x: JaxArray) -> JaxArray:
    """Flattens input tensor to a 2D tensor.

    Args:
        x: input tensor with dimensions (n_1, n_2, ..., n_k)

    Returns:
        The input tensor reshaped to two dimensions (n_1, n_sum),
        where n_sum is equal to the sum of n_2 to n_k.
    """
    return x.reshape([x.shape[0], -1])


def max_pool_2d(x: JaxArray,
                size: Union[Tuple[int, int], int] = 2,
                strides: Union[Tuple[int, int], int] = 2,
                padding: ConvPadding = ConvPadding.VALID) -> JaxArray:
    """Applies max pooling using a square 2D filter.

    Args:
        x: input tensor of shape NCHW.
        size: size of pooling filter.
        strides: stride step.
        padding: type of padding used in pooling operation.

    Returns:
        output tensor.
    """
    size = to_tuple(size, 2)
    strides = to_tuple(strides, 2)
    return lax.reduce_window(x, -jn.inf, lax.max, (1, 1) + size, (1, 1) + strides, padding=padding.value)


# Have to redefine relu since jnnf.relu isn't pickable.
def relu(x: JaxArray) -> JaxArray:
    """Rectified linear unit activation function.

    Args:
        x: input tensor.

    Returns:
        tensor with the element-wise output relu(x) = max(x, 0).
    """
    return jnnf.relu(x)


# Sample code for top_k with gradient support
# def top_k(x: jn.ndarray, k: int):
#     """Select the top k slices from the last dimension."""
#     bcast_idxs = jn.broadcast_to(jn.arange(x.shape[-1]), x.shape)
#     sorted_vals, sorted_idxs = lax.sort_key_val(x, bcast_idxs)
#     topk_vals = lax.slice_in_dim(sorted_vals, -k, sorted_vals.shape[-1], axis=-1)
#     topk_idxs = lax.slice_in_dim(sorted_idxs, -k, sorted_idxs.shape[-1], axis=-1)
#     topk_vals = jn.flip(topk_vals, axis=-1)
#     topk_idxs = jn.flip(topk_idxs, axis=-1)
#     return topk_vals, topk_idxs


def upscale(x: JaxArray) -> JaxArray:
    """Applies a 2x upscale of image batches of shape (N, C, H, W).

    Args:
        x: input tensor of shape NCHW.

    Returns:
        Input tensor where each each is upscaled by 2x.
    """
    s = x.shape
    x = x.reshape(s[:2] + (s[2], 1, s[3], 1))
    x = jn.tile(x, (1, 1, 1, 2, 1, 2))
    return x.reshape(s[:2] + (2 * s[2], 2 * s[3]))



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains type declarations for Objax."""

__all__ = ['FileOrStr', 'JaxArray', 'JaxDType']

from typing import Union, IO, BinaryIO

import jax.numpy as jn
from jax.interpreters.pxla import ShardedDeviceArray

FileOrStr = Union[str, IO[BinaryIO]]
JaxArray = Union[jn.ndarray, jn.DeviceArray, ShardedDeviceArray]
JaxDType = Union[jn.complex64, jn.complex128, jn.bfloat16,
                 jn.float16, jn.float32, jn.float64,
                 jn.int8, jn.int16, jn.int32, jn.int64,
                 jn.uint8, jn.uint16, jn.uint32, jn.uint64]



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re

from pkg_resources import parse_requirements
from setuptools import find_packages, setup

README_FILE = 'README.md'
REQUIREMENTS_FILE = 'requirements.txt'
VERSION_FILE = 'objax/_version.py'
VERSION_REGEXP = r'^__version__ = \'(\d+\.\d+\.\d+)\''

r = re.search(VERSION_REGEXP, open(VERSION_FILE).read(), re.M)
if r is None:
    raise RuntimeError(f'Unable to find version string in {VERSION_FILE}.')

version = r.group(1)
long_description = open(README_FILE, encoding='utf-8').read()
install_requires = [str(r) for r in parse_requirements(open(REQUIREMENTS_FILE, 'rt'))]

setup(
    name='objax',
    version=version,
    description='Objax is a machine learning framework that provides an Object Oriented layer for JAX.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Objax team',
    author_email='objax-dev@google.com',
    url='https://github.com/google/objax',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    install_requires=install_requires,
)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module with WideResNet implementation.

See https://arxiv.org/abs/1605.07146 for detail.
"""

__all__ = ['WRNBlock', 'WideResNetGeneral', 'WideResNet']

import functools
from typing import Callable, List

import objax
from objax.typing import JaxArray

BN_MOM = 0.9
BN_EPS = 1e-5


def conv_args(kernel_size: int, nout: int):
    """Returns list of arguments which are common to all convolutions.

    Args:
        kernel_size: size of convolution kernel (single number).
        nout: number of output filters.

    Returns:
        Dictionary with common convoltion arguments.
    """
    stddev = objax.functional.rsqrt(0.5 * kernel_size * kernel_size * nout)
    return dict(w_init=functools.partial(objax.random.normal, stddev=stddev),
                use_bias=False,
                padding=objax.constants.ConvPadding.SAME)


class WRNBlock(objax.Module):
    """WideResNet block."""

    def __init__(self,
                 nin: int,
                 nout: int,
                 stride: int = 1,
                 bn: Callable = functools.partial(objax.nn.BatchNorm2D, momentum=BN_MOM, eps=BN_EPS)):
        """Creates WRNBlock instance.

        Args:
            nin: number of input filters.
            nout: number of output filters.
            stride: stride for convolution and projection convolution in this block.
            bn: module which used as batch norm function.
        """
        if nin != nout or stride > 1:
            self.proj_conv = objax.nn.Conv2D(nin, nout, 1, strides=stride, **conv_args(1, nout))
        else:
            self.proj_conv = None

        self.norm_1 = bn(nin)
        self.conv_1 = objax.nn.Conv2D(nin, nout, 3, strides=stride, **conv_args(3, nout))
        self.norm_2 = bn(nout)
        self.conv_2 = objax.nn.Conv2D(nout, nout, 3, strides=1, **conv_args(3, nout))

    def __call__(self, x: JaxArray, training: bool) -> JaxArray:
        o1 = objax.functional.relu(self.norm_1(x, training))
        y = self.conv_1(o1)
        o2 = objax.functional.relu(self.norm_2(y, training))
        z = self.conv_2(o2)
        return z + self.proj_conv(o1) if self.proj_conv else z + x


class WideResNetGeneral(objax.nn.Sequential):
    """Base WideResNet implementation."""

    @staticmethod
    def mean_reduce(x: JaxArray) -> JaxArray:
        return x.mean((2, 3))

    def __init__(self,
                 nin: int,
                 nclass: int,
                 blocks_per_group: List[int],
                 width: int,
                 bn: Callable = functools.partial(objax.nn.BatchNorm2D, momentum=BN_MOM, eps=BN_EPS)):
        """Creates WideResNetGeneral instance.

        Args:
            nin: number of channels in the input image.
            nclass: number of output classes.
            blocks_per_group: number of blocks in each block group.
            width: multiplier to the number of convolution filters.
            bn: module which used as batch norm function.
        """
        widths = [int(v * width) for v in [16 * (2 ** i) for i in range(len(blocks_per_group))]]

        n = 16
        ops = [objax.nn.Conv2D(nin, n, 3, **conv_args(3, n))]
        for i, (block, width) in enumerate(zip(blocks_per_group, widths)):
            stride = 2 if i > 0 else 1
            ops.append(WRNBlock(n, width, stride, bn))
            for b in range(1, block):
                ops.append(WRNBlock(width, width, 1, bn))
            n = width
        ops += [bn(n),
                objax.functional.relu,
                self.mean_reduce,
                objax.nn.Linear(n, nclass, w_init=objax.nn.init.xavier_truncated_normal)
                ]
        super().__init__(ops)


class WideResNet(WideResNetGeneral):
    """WideResNet implementation with 3 groups.

    Reference:
        http://arxiv.org/abs/1605.07146
        https://github.com/szagoruyko/wide-residual-networks
    """

    def __init__(self,
                 nin: int,
                 nclass: int,
                 depth: int = 28,
                 width: int = 2,
                 bn: Callable = functools.partial(objax.nn.BatchNorm2D, momentum=BN_MOM, eps=BN_EPS)):
        """Creates WideResNet instance.

        Args:
            nin: number of channels in the input image.
            nclass: number of output classes.
            depth: number of convolution layers. (depth-4) should be divisible by 6
            width: multiplier to the number of convolution filters.
            bn: module which used as batch norm function.
        """
        assert (depth - 4) % 6 == 0, 'depth should be 6n+4'
        n = (depth - 4) // 6
        blocks_per_group = [n] * 3
        super().__init__(nin, nclass, blocks_per_group, width, bn)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax

import objax
from objax.typing import JaxArray


class ConvNet(objax.nn.Sequential):
    """ConvNet implementation."""

    @staticmethod
    def _mean_reduce(x: JaxArray) -> JaxArray:
        return x.mean((2, 3))

    def __init__(self, nin, nclass, scales, filters, filters_max,
                 pooling=objax.functional.max_pool_2d, **kwargs):
        """Creates ConvNet instance.

        Args:
            nin: number of channels in the input image.
            nclass: number of output classes.
            scales: number of pooling layers, each of which reduces spatial dimension by 2.
            filters: base number of convolution filters.
                     Number of convolution filters is increased by 2 every scale until it reaches filters_max.
            filters_max: maximum number of filters.
            pooling: type of pooling layer.
        """
        del kwargs

        def nf(scale):
            return min(filters_max, filters << scale)

        ops = [objax.nn.Conv2D(nin, nf(0), 3), objax.functional.leaky_relu]
        for i in range(scales):
            ops.extend([objax.nn.Conv2D(nf(i), nf(i), 3), objax.functional.leaky_relu,
                        objax.nn.Conv2D(nf(i), nf(i + 1), 3), objax.functional.leaky_relu,
                        jax.partial(pooling, size=2, strides=2)])
        ops.extend([objax.nn.Conv2D(nf(scales), nclass, 3), self._mean_reduce])
        super().__init__(ops)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from . import functional
from . import io
from . import jaxboard
from . import nn
from . import optimizer
from . import privacy
from . import random
from . import typing
from . import util
from ._version import __version__
from .constants import *
from .gradient import *
from .module import *
from .variable import *

assert sys.version_info >= (3, 6)
